import logging
import os
import time
from typing import Sequence

from hhd.plugins import Config, Context, Event, HHDPlugin, load_relative_yaml

from adjustor.core.platform import set_platform_profile
from adjustor.core.const import DeviceTDPv2
from adjustor.i18n import _

logger = logging.getLogger(__name__)

APPLY_DELAY = 0.7
TDP_DELAY = 0.1
SLEEP_DELAY = 4

TDP_PL3_FN = "/sys/class/firmware-attributes/msi-wmi-platform/attributes/ppt_pl3_fppt/current_value"
TDP_PL2_FN = "/sys/class/firmware-attributes/msi-wmi-platform/attributes/ppt_pl2_sppt/current_value"
TDP_PL1_FN = "/sys/class/firmware-attributes/msi-wmi-platform/attributes/ppt_pl1_spl/current_value"

FAN_CURVE_ENDPOINT = "/sys/class/hwmon"
FAN_CURVE_NAME = "msi_wmi_platform"

# Default MSI Claw AI+ 8 curve is the following
# [0  50  60  70  80  88] temp (C)
# [0  40  49  58  67  75] pwm  (%)
# [0 102 124 147 170 191] / 2.55


# Claw 1st gen
# DC values
#       PL1 PL2
# perf   43  45
# bal    35  35
# power  20  20
#
# AC values
#       PL1 PL2
# perf   35  35
# bal    30  30
# power  20  20

# Claw AI+
# DC values
#       PL1 PL2
# perf   30  37
# bal    12  37
# power   8  37
#
# AC values
#       PL1 PL2
# perf   30  37
# bal    12  37
# power   8  37
#
# Claw 8A
# Both AC/DC values
#       SPL SPPT FPPT
# perf   28   45   55
# bal    20   33   43
# power  15   28   33

POINTS = [0, 50, 60, 70, 80, 88]
DEFAULT_CURVE = [0, 102, 124, 147, 170, 191]


def set_tdp(pretty: str, fn: str, val: int):
    logger.info(f"Setting tdp value '{pretty}' to {val} by writing to:\n{fn}")
    try:
        with open(fn, "w") as f:
            f.write(f"{val}\n")
        return True
    except Exception as e:
        logger.error(f"Failed writing value with error:\n{e}")
        return False


def find_fan_curve_dir():
    for dir in os.listdir(FAN_CURVE_ENDPOINT):
        name_fn = os.path.join(FAN_CURVE_ENDPOINT, dir, "name")
        with open(name_fn, "r") as f:
            name = f.read().strip()
        if name == FAN_CURVE_NAME:
            return os.path.join(FAN_CURVE_ENDPOINT, dir)
    return None


def set_fan_curve(points: list[int], curve: list[int]):
    point_str = ",".join([f"{p:> 4d} C" for p in points])
    curve_str = ",".join([f"{p:> 4d} /" for p in curve])
    logger.info(f"Setting the following fan curve:\n{point_str}\n{curve_str} 255")

    dir = find_fan_curve_dir()
    if not dir:
        logger.error(f"Could not find hwmon with name:\n'{FAN_CURVE_NAME}'")
        return False

    for fan in (1, 2):
        for i, (temp, speed) in enumerate(zip(points, curve)):
            with open(os.path.join(dir, f"pwm{fan}_auto_point{i+1}_temp"), "w") as f:
                f.write(f"{temp}")
            with open(os.path.join(dir, f"pwm{fan}_auto_point{i+1}_pwm"), "w") as f:
                f.write(f"{speed}")

    for fan in (1, 2):
        with open(os.path.join(dir, f"pwm{fan}_enable"), "w") as f:
            f.write(f"1")
        if fan == 1:
            time.sleep(TDP_DELAY)

    return True


def disable_fan_curve():
    logger.info(f"Disabling custom fan curve.")

    dir = find_fan_curve_dir()
    if not dir:
        logger.error(f"Could not find hwmon with name:\n'{FAN_CURVE_NAME}'")
        return False

    for fan in (1, 2):
        with open(os.path.join(dir, f"pwm{fan}_enable"), "w") as f:
            f.write(f"2")
        if fan == 1:
            time.sleep(TDP_DELAY)

    return True


def has_msi_driver():
    return os.path.exists(TDP_PL1_FN) and os.path.exists(TDP_PL2_FN)


class MsiDriverPlugin(HHDPlugin):
    def __init__(self, tdp_data: DeviceTDPv2) -> None:
        self.name = f"adjustor_msi"
        self.priority = 6
        self.log = "adjm"
        self.enabled = False
        self.initialized = False
        self.enforce_limits = True
        self.startup = True
        self.old_conf = None
        self.mode = None
        self.cycle_tdp = None

        self.extreme_standby = None
        self.extreme_supported = None

        self.queue_fan = None
        self.queue_tdp = None
        self.new_tdp = None
        self.new_mode = None
        self.old_target = None
        self.pp = None
        self.sys_tdp = False
        self.tdp_data = tdp_data

    def settings(self):
        if not self.enabled:
            self.initialized = False
            self.old_conf = None
            self.startup = True
            return {}

        if not has_msi_driver():
            logger.error("MSI driver not found. Disabling TDP controls.")
            self.initialized = False

            return {"tdp": {"msi": load_relative_yaml("error.yml")}}

        self.initialized = True
        out = {"tdp": {"msi": load_relative_yaml("settings.yml")}}

        # Set units
        out["tdp"]["msi"]["children"]["tdp"]["modes"]["quiet"][
            "unit"
        ] = f"{self.tdp_data['quiet'][0]}W"
        out["tdp"]["msi"]["children"]["tdp"]["modes"]["balanced"][
            "unit"
        ] = f"{self.tdp_data['balanced'][0]}W"

        # Set performance pretty print
        if (
            self.tdp_data.get("performance_dc", None)
            and self.tdp_data["performance_dc"] != self.tdp_data["performance"][0]
        ):
            perf_tdp = (
                f"{self.tdp_data['performance_dc']}W/{self.tdp_data['performance'][0]}W"
            )
        else:
            perf_tdp = f"{self.tdp_data['performance'][0]}W"
        out["tdp"]["msi"]["children"]["tdp"]["modes"]["performance"]["unit"] = perf_tdp

        # Set custom pretty print
        if (
            self.tdp_data.get("max_tdp_dc", None)
            and self.tdp_data["max_tdp_dc"] != self.tdp_data["max_tdp"]
        ):
            custom_tdp = f"→ {self.tdp_data['max_tdp_dc']}W/{self.tdp_data['max_tdp']}W"
        else:
            custom_tdp = f"→ {self.tdp_data['max_tdp']}W"
        out["tdp"]["msi"]["children"]["tdp"]["modes"]["custom"]["unit"] = custom_tdp

        # Fix custom slider
        custom_sel = out["tdp"]["msi"]["children"]["tdp"]["modes"]["custom"][
            "children"
        ]["tdp"]
        custom_sel["min"] = self.tdp_data["min_tdp"]
        custom_sel["max"] = self.tdp_data["max_tdp"]
        custom_sel["default"] = self.tdp_data["balanced"][0]

        # Add overclocking
        if not self.enforce_limits and self.tdp_data.get("max_tdp_oc", None):
            out["tdp"]["msi"]["children"]["tdp"]["modes"]["custom"]["children"]["tdp"][
                "max"
            ] = self.tdp_data["max_tdp_oc"]

        # Remove cycle for laptops (for now)
        if not self.tdp_data.get("supports_cycle", None):
            del out["tdp"]["msi"]["children"]["cycle_tdp"]

        return out

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit

        # TODO: Remove me once the autoload issue is fixed
        os.system("modprobe msi_wmi_platform")

    def update(self, conf: Config):
        self.enabled = conf["hhd.settings.tdp_ready"].to(bool)
        new_enforce_limits = conf["hhd.settings.enforce_limits"].to(bool)
        new_lims = new_enforce_limits != self.enforce_limits
        self.enforce_limits = new_enforce_limits

        if not self.enabled or not self.initialized or new_lims:
            self.old_conf = None
            self.startup = True
            return

        # If not old config, exit, as values can not be set
        if not self.old_conf:
            self.old_conf = conf["tdp.msi"]
            return

        curr = time.perf_counter()

        #
        # TDP
        #
        new_target = None
        new_tdp = self.new_tdp
        self.new_tdp = None
        new_mode = self.new_mode
        self.new_mode = None
        if new_tdp:
            mode = "custom"
            conf["tdp.msi.tdp.mode"] = mode
        elif new_mode:
            mode = new_mode
            conf["tdp.msi.tdp.mode"] = mode
        else:
            mode = conf["tdp.msi.tdp.mode"].to(str)
        self.mode = mode

        tdp_reset = False
        if mode is not None and mode != self.old_conf["tdp.mode"].to(str):
            if not new_tdp:
                self.sys_tdp = False
            tdp_reset = True

        if mode is not None and self.startup:
            tdp_reset = True

        # Handle EPP for presets
        if tdp_reset and mode != "custom":
            match mode:
                case "quiet":
                    set_platform_profile("low-power")
                    spl, sppt, fppt = self.tdp_data["quiet"]
                    set_tdp("pl1_spl", TDP_PL1_FN, spl)
                    set_tdp("pl2_sppt", TDP_PL2_FN, sppt)
                    if fppt is not None:
                        set_tdp("pl3_fppt", TDP_PL3_FN, fppt)
                    new_target = "power"
                case "balanced":
                    set_platform_profile("balanced")
                    spl, sppt, fppt = self.tdp_data["balanced"]
                    set_tdp("pl1_spl", TDP_PL1_FN, spl)
                    set_tdp("pl2_sppt", TDP_PL2_FN, sppt)
                    if fppt is not None:
                        set_tdp("pl3_fppt", TDP_PL3_FN, fppt)
                    new_target = "balanced"
                case _:  # "performance":
                    set_platform_profile("performance")
                    spl, sppt, fppt = self.tdp_data["performance"]
                    set_tdp("pl1_spl", TDP_PL1_FN, spl)
                    set_tdp("pl2_sppt", TDP_PL2_FN, sppt)
                    if fppt is not None:
                        set_tdp("pl3_fppt", TDP_PL3_FN, fppt)
                    new_target = "performance"

        # In custom mode, re-apply settings with debounce
        tdp_set = False
        if mode == "custom":
            # Check user changed values
            if new_tdp:
                steady = new_tdp
                conf["tdp.msi.tdp.custom.tdp"] = steady
            else:
                steady = conf["tdp.msi.tdp.custom.tdp"].to(int)

            if self.enforce_limits:
                if steady < self.tdp_data["min_tdp"]:
                    steady = self.tdp_data["min_tdp"]
                    conf["tdp.msi.tdp.custom.tdp"] = steady
                elif steady > self.tdp_data["max_tdp"]:
                    steady = self.tdp_data["max_tdp"]
                    conf["tdp.msi.tdp.custom.tdp"] = steady

            steady_updated = steady and steady != self.old_conf["tdp.custom.tdp"].to(
                int
            )
            if not new_tdp and steady_updated:
                self.sys_tdp = False

            steady_updated |= tdp_reset

            if self.startup and (
                steady > self.tdp_data["max_tdp"] or steady < self.tdp_data["min_tdp"]
            ):
                logger.warning(
                    f"TDP ({steady}) outside the device spec. Resetting for stability reasons."
                )
                steady = min(
                    max(steady, self.tdp_data["min_tdp"]), self.tdp_data["max_tdp"]
                )
                conf["tdp.msi.tdp.custom.tdp"] = steady
                steady_updated = True

            boost = conf["tdp.msi.tdp.custom.boost"].to(bool)
            boost_updated = boost != self.old_conf["tdp.custom.boost"].to(bool)

            # If yes, queue an update
            # Debounce
            if self.startup or steady_updated or boost_updated:
                self.queue_tdp = curr + APPLY_DELAY

            tdp_set = self.queue_tdp and self.queue_tdp < curr
            if tdp_set:
                if steady < self.tdp_data["min_tdp"]:
                    steady = 5
                if steady < self.tdp_data["balanced_min"]:
                    set_platform_profile("low-power")
                    new_target = "power"
                elif steady < self.tdp_data["performance_min"]:
                    set_platform_profile("balanced")
                    new_target = "balanced"
                else:
                    set_platform_profile("performance")
                    new_target = "performance"

                self.queue_tdp = None
                if boost:
                    if os.path.exists(TDP_PL3_FN):
                        time.sleep(TDP_DELAY)
                        set_tdp(
                            "fast",
                            TDP_PL3_FN,
                            min(
                                max(
                                    steady,
                                    self.tdp_data["max_tdp_fppt"]
                                    or self.tdp_data["max_tdp"],
                                ),
                                (
                                    int(
                                        steady
                                        * self.tdp_data["max_tdp_fppt"]
                                        / self.tdp_data["max_tdp"]
                                    )
                                    if self.tdp_data["max_tdp_fppt"]
                                    else steady
                                ),
                            ),
                        )
                    time.sleep(TDP_DELAY)
                    set_tdp(
                        "slow",
                        TDP_PL2_FN,
                        min(
                            max(
                                steady,
                                self.tdp_data["max_tdp_sppt"]
                                or self.tdp_data["max_tdp"],
                            ),
                            (
                                int(
                                    steady
                                    * self.tdp_data["max_tdp_sppt"]
                                    / self.tdp_data["max_tdp"]
                                )
                                if self.tdp_data["max_tdp_sppt"]
                                else steady
                            ),
                        ),
                    )
                    time.sleep(TDP_DELAY)
                    set_tdp("steady", TDP_PL1_FN, steady)
                else:
                    if os.path.exists(TDP_PL3_FN):
                        time.sleep(TDP_DELAY)
                        set_tdp("fast", TDP_PL3_FN, steady)
                    time.sleep(TDP_DELAY)
                    set_tdp("slow", TDP_PL2_FN, steady)
                    time.sleep(TDP_DELAY)
                    set_tdp("steady", TDP_PL1_FN, steady)

        if new_target and new_target != self.old_target:
            self.old_target = new_target
            self.emit({"type": "energy", "status": new_target})
            self.pp = new_target

        # Handle fan curve resets
        if conf["tdp.msi.fan.manual.reset"].to(bool):
            conf["tdp.msi.fan.manual.reset"] = False
            for k, v in zip(POINTS, DEFAULT_CURVE):
                conf[f"tdp.msi.fan.manual.st{k}"] = v

        manual_fan_curve = conf["tdp.msi.fan.mode"].to(str) == "manual"

        # Check if fan curve has changed
        # Use debounce logic on these changes
        if ((tdp_reset and mode != "custom") or tdp_set) and manual_fan_curve:
            self.queue_fan = curr + APPLY_DELAY

        for i in POINTS:
            if conf[f"tdp.msi.fan.manual.st{i}"].to(int) != self.old_conf[
                f"fan.manual.st{i}"
            ].to(int):
                self.queue_fan = curr + APPLY_DELAY
        # If mode changes, only apply curve if set to manual
        # otherwise disable and reset tdp
        if conf["tdp.msi.fan.mode"].to(str) != self.old_conf["fan.mode"].to(str):
            if conf["tdp.msi.fan.mode"].to(str) == "manual":
                self.queue_fan = curr + APPLY_DELAY
            else:
                try:
                    disable_fan_curve()
                except Exception as e:
                    logger.error(f"Could not disable fan curve. Error:\n{e}")
                self.queue_tdp = curr + APPLY_DELAY

        apply_curve = self.queue_fan and self.queue_fan < curr
        if apply_curve or tdp_set:
            try:
                if conf["tdp.msi.fan.mode"].to(str) == "manual":
                    set_fan_curve(
                        POINTS,
                        [
                            min(
                                int(conf[f"tdp.msi.fan.manual.st{i}"].to(int) * 2.55),
                                255,
                            )
                            for i in POINTS
                        ],
                    )
            except Exception as e:
                logger.error(f"Could not set fan curve. Error:\n{e}")
            self.queue_fan = None

        # Show steam message
        if self.sys_tdp:
            conf["tdp.msi.sys_tdp"] = _("Steam is controlling TDP")
        else:
            conf["tdp.msi.sys_tdp"] = ""

        # Save current config
        self.cycle_tdp = self.tdp_data.get("supports_cycle", False) and conf.get(
            "tdp.msi.cycle_tdp", False
        )
        self.old_conf = conf["tdp.msi"]

        if self.startup:
            self.startup = False

    def notify(self, events: Sequence[Event]):
        for ev in events:
            if ev["type"] == "tdp":
                self.new_tdp = ev["tdp"]
                self.sys_tdp = ev["tdp"] is not None
            elif ev["type"] == "ppd":
                match ev["status"]:
                    case "power":
                        self.new_mode = "quiet"
                    case "balanced":
                        self.new_mode = "balanced"
                    case "performance":
                        self.new_mode = "performance"
            elif ev["type"] == "special" and ev.get("event", None) == "wakeup":
                logger.info(
                    f"Waking up from sleep, resetting TDP after {SLEEP_DELAY} seconds."
                )
                self.queue_tdp = time.time() + SLEEP_DELAY
            elif (
                ev["type"] == "acpi"
                and ev["event"] in ("ac", "dc")
                and not self.queue_tdp
            ):
                logger.info(
                    f"Power adapter status switched to '{ev['event']}', resetting TDP."
                )
                self.queue_tdp = time.time() + APPLY_DELAY
            elif (
                self.cycle_tdp
                and ev["type"] == "special"
                and ev["event"] == "xbox_y_internal"
            ) or (ev["type"] == "special" and ev["event"] == "tdp_cycle"):
                match self.mode:
                    case "quiet":
                        self.new_mode = "balanced"
                        event = "tdp_cycle_balanced"
                    case "balanced":
                        self.new_mode = "performance"
                        event = "tdp_cycle_performance"
                    case "performance":
                        self.new_mode = "custom"
                        event = "tdp_cycle_custom"
                    case "custom":
                        self.new_mode = "quiet"
                        event = "tdp_cycle_quiet"
                    case _:
                        self.new_mode = "balanced"
                        event = "tdp_cycle_balanced"

                logger.info(f"Cycling TDP to '{self.new_mode}'")
                if self.emit:
                    self.emit({"type": "special", "event": event})

    def close(self):
        pass



import os

FAN_HWMONS_LEGACY = ["oxpec"]
FAN_HWMONS = ["oxp_ec", "gpdfan", "ayaneo_ec"]
HWMON_DIR = "/sys/class/hwmon"


def get_hwmon():
    for dir in os.listdir(HWMON_DIR):
        if dir.startswith("hwmon"):
            yield dir


def find_edge_temp():
    for hwmon in get_hwmon():
        with open(f"{HWMON_DIR}/{hwmon}/name") as f:
            name = f.read().strip()

        if name != "amdgpu":
            continue

        # For sanity, check the device has CPUs to avoid hooking an eGPU.
        if not os.path.exists(f"{HWMON_DIR}/{hwmon}/device/local_cpus"):
            continue

        if not os.path.exists(f"{HWMON_DIR}/{hwmon}/temp1_input"):
            continue

        return f"{HWMON_DIR}/{hwmon}/temp1_input"


def find_tctl_temp():
    for hwmon in get_hwmon():
        with open(f"{HWMON_DIR}/{hwmon}/name") as f:
            name = f.read().strip()

        if name != "k10temp":
            continue

        # For sanity, check the device has CPUs to avoid hooking an eGPU.
        if not os.path.exists(f"{HWMON_DIR}/{hwmon}/device/local_cpus"):
            continue

        if not os.path.exists(f"{HWMON_DIR}/{hwmon}/temp1_input"):
            continue

        return f"{HWMON_DIR}/{hwmon}/temp1_input"


def find_fans():
    """Finds tunable fans with endpoints pwmX and pwmX_enable."""
    fans = []
    for hwmon in get_hwmon():
        with open(f"{HWMON_DIR}/{hwmon}/name") as f:
            name = f.read().strip()

        if name not in FAN_HWMONS and name not in FAN_HWMONS_LEGACY:
            continue

        for fn in os.listdir(f"{HWMON_DIR}/{hwmon}"):
            if (
                fn.startswith("pwm")
                and fn[3:].isdigit()
                and os.path.exists(f"{HWMON_DIR}/{hwmon}/{fn}_enable")
            ):
                idx = fn[3:]
                speed = f"fan{idx}_input"
                if speed in os.listdir(f"{HWMON_DIR}/{hwmon}"):
                    speed_fn = f"{HWMON_DIR}/{hwmon}/{speed}"
                else:
                    speed_fn = None
                fans.append(
                    (
                        f"{HWMON_DIR}/{hwmon}/{fn}",
                        f"{HWMON_DIR}/{hwmon}/{fn}_enable",
                        speed_fn,
                        name in FAN_HWMONS_LEGACY,
                    )
                )

    return fans


def read_temp(path: str) -> float:
    with open(path, "r") as f:
        return int(f.read()) / 1000


def read_fan_speed(path: str) -> int:
    with open(path, "r") as f:
        return int(f.read())


def write_fan_speed(path: str, speed: int):
    with open(path, "w") as f:
        f.write(str(speed))



import logging
import os
from threading import Event as TEvent
from threading import Thread
from typing import Sequence

from hhd.plugins import Context, HHDPlugin, HHDSettings, load_relative_yaml
from hhd.plugins.conf import Config
from hhd.plugins.plugin import Emitter

from adjustor.core.acpi import check_perms, initialize
from adjustor.core.const import CPU_DATA, DEV_DATA, ASUS_DATA, MSI_DATA

from .i18n import _

logger = logging.getLogger(__name__)

CONFLICTING_PLUGINS = {
    "SimpleDeckyTDP": "homebrew/plugins/SimpleDeckyTDP",
    "PowerControl": "homebrew/plugins/PowerControl",
}


class AdjustorInitPlugin(HHDPlugin):
    def __init__(
        self,
        min_tdp: int,
        default_tdp: int,
        max_tdp: int,
        use_acpi_call: bool = True,
    ) -> None:
        self.name = f"adjustor_init"
        self.priority = 5
        self.log = "adjs"
        self.init = False
        self.failed = False
        self.has_decky = False
        self.enabled = False
        self.action_enabled = False
        self.use_acpi_call = use_acpi_call

        self.enabled = False
        self.enfoce_limits = True

        self.t = None
        self.should_exit = None

        self.min_tdp = min_tdp
        self.default_tdp = default_tdp
        self.max_tdp = max_tdp
        self.tdp_set = None

    def open(self, emit: Emitter, context: Context):
        self.context = context
        self.emit = emit

    def settings(self):
        v = load_relative_yaml("settings.yml")
        sets = {
            "hhd": {"settings": v["hhd"], "steamos": v["steamos"]},
        }
        if os.environ.get("HHD_ADJ_ENABLE_TDP"):
            sets["hhd"]["settings"]["children"]["tdp_enable"]["default"] = True

        if self.enabled and not self.failed:
            self.action_enabled = False
        else:
            self.action_enabled = True
            sets["tdp"] = {"tdp": v["tdp"]}
            if not self.has_decky:
                del sets["tdp"]["tdp"]["children"]["decky_info"]
                del sets["tdp"]["tdp"]["children"]["decky_remove"]

        return sets

    def update(self, conf: Config):
        if (
            self.action_enabled
            and self.has_decky
            and conf["tdp.tdp.decky_remove"].to(bool)
        ):
            # Preparation
            logger.warning("Removing Decky plugins")
            conf["tdp.tdp.decky_remove"] = False
            conf["hhd.settings.tdp_enable"] = True
            self.has_decky = False
            self.failed = False

            for usr in os.listdir("/home"):
                move_path = os.path.join("/home", usr, "homebrew/plugins/hhd-disabled")
                if os.path.exists(move_path):
                    logger.warning(f"Removing old backup path: '{move_path}'")
                    os.system(f"rm -rf {move_path}")
                os.makedirs(
                    move_path,
                    exist_ok=True,
                )

                logger.warning("Stopping Decky.")
                try:
                    os.system("systemctl stop plugin_loader")
                except Exception as e:
                    logger.error(f"Failed to restart Decky:\n{e}")

                for name, ppath in CONFLICTING_PLUGINS.items():
                    path = os.path.join("/home", usr, ppath)
                    if os.path.exists(path):
                        new_path = os.path.join(move_path, name)
                        logger.warning(
                            f"Moving plugin '{name}' from:\n{path}\nto:\n{new_path}"
                        )
                        os.rename(path, new_path)

                logger.warning("Restarting Decky.")
                try:
                    os.system("systemctl start plugin_loader")
                except Exception as e:
                    logger.error(f"Failed to restart Decky:\n{e}")

            # TDP controls are already enabled.
            logger.warning(f"Enabling TDP controls.")
        
        if self.action_enabled and conf.get_action("tdp.tdp.tdp_enable"):
            conf["hhd.settings.tdp_enable"] = True
            self.failed = False

        enabled = conf.get("hhd.settings.tdp_enable", False)
        
        if not enabled:
            conf["hhd.settings.tdp_ready"] = False
            conf["hhd.steamos.tdp_status"] = "disabled"
            conf["hhd.steamos.tdp_min"] = None
            conf["hhd.steamos.tdp_default"] = None
            conf["hhd.steamos.tdp_max"] = None
            self.init = False

        if self.enabled != enabled:
            self.emit({"type": "settings"})
            if not enabled:
                self.enabled = False
                self._stop()
        self.enabled = enabled

        if not enabled:
            return
        
        new_enforce_limits = conf["hhd.settings.enforce_limits"].to(bool)
        if new_enforce_limits != self.enfoce_limits:
            self.emit({"type": "settings"})
        self.enfoce_limits = new_enforce_limits

        # If steam sets a value other than max TDP, assume TDP slider
        # is enabled. Then, disable it again when max TDP is set but
        # allow max TDP to be set normally to reset it.
        if self.tdp_set is not None:
            conf["hhd.steamos.tdp_set"] = self.tdp_set

        if self.init or not enabled or self.failed:
            return

        for usr in os.listdir("/home"):
            for name, path in CONFLICTING_PLUGINS.items():
                if os.path.exists(os.path.join(usr, path)):
                    err = f'Found "{name}" at:\n{path}\n' + _(
                        "Disable Decky TDP plugins using the button below to continue."
                    )
                    self.emit({"type": "settings"})
                    self.has_decky = True
                    conf["tdp.tdp.tdp_error"] = err
                    conf["hhd.settings.tdp_ready"] = False
                    conf["hhd.steamos.tdp_status"] = "conflict"
                    logger.error(err)
                    self.failed = True
                    self.enabled = False
                    self._stop()
                    return

        if self.use_acpi_call:
            initialize()
            if not check_perms():
                conf["hhd.settings.tdp_ready"] = False
                conf["tdp.tdp.tdp_error"] = (
                    "Can not write to 'acpi_call'. It is required for TDP."
                )
                self.failed = True
                self.enabled = False
                self._stop()
                return
        
        self._start()
        self.failed = False
        self.init = True
        conf["hhd.steamos.tdp_status"] = "enabled"
        conf["hhd.steamos.tdp_min"] = self.min_tdp
        conf["hhd.steamos.tdp_default"] = self.default_tdp
        conf["hhd.steamos.tdp_max"] = self.max_tdp
        conf["hhd.settings.tdp_ready"] = True
        conf["tdp.tdp.tdp_error"] = ""

    def _start(self):
        if self.should_exit:
            return
        self.should_exit = TEvent()
        if not self.t:
            try:
                from .events import loop_process_events

                self.t = Thread(
                    target=loop_process_events, args=(self.emit, self.should_exit)
                )
                self.t.start()
            except Exception as e:
                logger.warning(
                    f"Could not init ACPI event handling. Is pyroute2 installed?"
                )

    def _stop(self):
        if not self.should_exit:
            return
        self.should_exit.set()
        if self.t:
            self.t.join()
            self.t = None
        self.should_exit = None

    def close(self):
        self._stop()
    
    def notify(self, events: Sequence):
        for ev in events:
            if ev["type"] == "tdp":
                self.tdp_set = ev["tdp"] != self.max_tdp

LEGION_GO_DMIS = ["83E1", "83N0", "83N1"]
LEGION_GO_S_DMIS = ["83L3", "83N6", "83Q2", "83Q3"]


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if os.environ.get("HHD_ADJ_DISABLE"):
        return []

    if len(existing):
        return existing

    from .drivers.asus import AsusDriverPlugin
    from .drivers.lenovo import LenovoDriverPlugin
    from .drivers.msi import MsiDriverPlugin
    from .drivers.smu import SmuDriverPlugin, SmuQamPlugin
    from .drivers.gpu import GpuPlugin
    from .drivers.battery import BatteryPlugin

    drivers = []
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        prod = f.read().strip()
    with open("/sys/devices/virtual/dmi/id/board_name") as f:
        board = f.read().strip()
    with open("/proc/cpuinfo") as f:
        cpuinfo = f.read().strip()

    use_acpi_call = False
    drivers_matched = False

    # FIXME: Switch to per device
    # But all devices use the same values
    # pretty much
    min_tdp = 4
    default_tdp = 15
    max_tdp = 30

    legion_s = prod in LEGION_GO_S_DMIS
    legion_t = prod in LEGION_GO_DMIS
    if (legion_t or legion_s) and not bool(os.environ.get("HHD_ADJ_ALLY")):
        drivers.append(LenovoDriverPlugin(legion_s=legion_s))
        drivers_matched = True
        use_acpi_call = True
    if legion_s:
        max_tdp = 33

    for k, v in ASUS_DATA.items():
        if k in prod:
            drivers.append(AsusDriverPlugin(v))
            drivers_matched = True
            min_tdp = v["min_tdp"]
            max_tdp = v["max_tdp"]
            break

    for k, v in MSI_DATA.items():
        if k in board:
            drivers.append(MsiDriverPlugin(v))
            drivers_matched = True
            min_tdp = v["min_tdp"]
            max_tdp = v["max_tdp"]
            break

    if os.environ.get("HHD_ADJ_DEBUG") or os.environ.get("HHD_ENABLE_SMU"):
        drivers_matched = False

    if not drivers_matched and prod in DEV_DATA:
        dev, cpu, pp_enable, energy_map = DEV_DATA[prod]

        try:
            # Set values for the steam slider
            if dev["skin_limit"].smin:
                min_tdp = dev["skin_limit"].smin
            if dev["skin_limit"].default:
                default_tdp = dev["skin_limit"].default
            if dev["skin_limit"].smax:
                max_tdp = dev["skin_limit"].smax
        except Exception as e:
            logger.error(f"Failed to get TDP limits for {prod}:\n{e}")

        pp_enable |= bool(os.environ.get("HHD_ADJ_DEBUG"))
        drivers.append(
            SmuDriverPlugin(
                dev,
                cpu,
                platform_profile=pp_enable,
            )
        )
        drivers.append(
            SmuQamPlugin(
                dev,
                energy_map,
                pp_enable=pp_enable,
                init_tdp=not prod == "83E1",
            ),
        )
        drivers_matched = True
        use_acpi_call = True

    if not drivers_matched:
        for name, (dev, cpu, energy_map) in CPU_DATA.items():
            if name in cpuinfo:
                drivers.append(
                    SmuDriverPlugin(
                        dev,
                        cpu,
                        platform_profile=True,
                    )
                )
                drivers.append(
                    SmuQamPlugin(dev, energy_map),
                )
                use_acpi_call = True
                break

    if not drivers:
        from .drivers.general import GeneralPowerPlugin

        logger.info(f"No tdp drivers found for this device, using generic plugin.")

        return [GeneralPowerPlugin(), BatteryPlugin(always_enable=True)]

    return [
        *drivers,
        AdjustorInitPlugin(
            use_acpi_call=use_acpi_call,
            min_tdp=min_tdp,
            default_tdp=default_tdp,
            max_tdp=max_tdp,
        ),
        BatteryPlugin(),
        GpuPlugin(),
    ]



import logging
import os
import time

WAKEUP_DIR = "/sys/class/wakeup"
SMBIOS_FN = "/sys/firmware/dmi/entries/1-0/raw"
WAKE_COUNT_FN = "/sys/power/wakeup_count"

def configure_valid_wake_events():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
    )

    logger = logging.getLogger(__name__)

    for entry in os.scandir(WAKEUP_DIR):
        if not entry.is_dir():
            continue

        with open(os.path.join(entry.path, "name"), "r") as f:
            name = f.read().strip()

        try:
            with open(os.path.join(entry.path, "device/power/wakeup"), "r") as f:
                wakeup_en = f.read().strip()
        except FileNotFoundError:
            wakeup_en = "NA"

        if "PNP0C0" in name:
            continue

        if wakeup_en != "enabled":
            continue
        
        logger.info(f"Disabling {name} ({entry.path}) wake events")
        # with open(os.path.join(entry.path, "device/power/wakeup"), "w") as f:
        #     f.write("disabled")


def main_loop():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
    )

    logger = logging.getLogger(__name__)

    gwakeup_count = 0
    event_counts = {}
    wakeup_counts = {}
    old_reason = None

    while True:
        for entry in os.scandir(WAKEUP_DIR):
            if not entry.is_dir():
                continue

            with open(os.path.join(entry.path, "event_count"), "r") as f:
                event_count = f.read().strip()

            with open(os.path.join(entry.path, "wakeup_count"), "r") as f:
                wakeup_count = f.read().strip()

            try:
                with open(os.path.join(entry.path, "device/power/wakeup"), "r") as f:
                    wakeup_en = f.read().strip()
            except FileNotFoundError:
                wakeup_en = "NA"

            diff = False
            if entry.path not in wakeup_counts:
                wakeup_counts[entry.path] = wakeup_count
                # diff = True

            if wakeup_counts[entry.path] != wakeup_count:
                diff = True
                wakeup_counts[entry.path] = wakeup_count

            if entry.path not in event_counts:
                event_counts[entry.path] = event_count
                # diff = True

            if event_counts[entry.path] != event_count:
                diff = True
                event_counts[entry.path] = event_count

            if not diff:
                continue

            with open(os.path.join(entry.path, "name"), "r") as f:
                name = f.read().strip()

            wakeup_path = entry.path.split("/")[-1]

            logger.info(
                f"{name:>20s} ({wakeup_path:>8s}): enabled: {wakeup_en}, {wakeup_count:>3s} wakeups, {event_count:>3s} events"
            )
            wakeup_counts[entry.path] = wakeup_count

        with open(SMBIOS_FN, "rb") as f:
            smbios = f.read()

        with open(WAKE_COUNT_FN, "r") as f:
            wake_count = int(f.read().strip())

        wakeup_reason = smbios[24]
        # 00h Reserved
        # 01h Other
        # 02h Unknown
        # 03h APM Timer
        # 04h Modem Ring
        # 05h LAN Remote
        # 06h Power Switch
        # 07h PCI PME#
        # 08h AC Power Restored
        match wakeup_reason:
            case 0x00:
                reason = "Reserved"
            case 0x01:
                reason = "Other"
            case 0x02:
                reason = "Unknown"
            case 0x03:
                reason = "APM Timer"
            case 0x04:
                reason = "Modem Ring"
            case 0x05:
                reason = "LAN Remote"
            case 0x06:
                reason = "Power Switch"
            case 0x07:
                reason = "PCI PME#"
            case 0x08:
                reason = "AC Power Restored"
            case _:
                reason = f"Unknown ({wakeup_reason:02X})"

        if gwakeup_count != wake_count:
            logger.info(f"Wakeups {wake_count:5d}, reason: {reason}")
            gwakeup_count = wake_count

        time.sleep(0.2)


def main():
    try:
        main_loop()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



from .acpi import call, read
from typing import Sequence, Literal

import logging

logger = logging.getLogger(__name__)

MIN_CURVE = [44, 48, 55, 60, 71, 79, 87, 87, 100, 100]

TdpMode = Literal["quiet", "balanced", "performance", "custom"]


def get_fan_curve():
    logger.info("Retrieving fan curve.")
    o = call(r"\_SB.GZFD.WMAB", [0, 0x05, bytes([0, 0, 0, 0])], risky=False)
    if not o:
        return None
    o = read()
    if not isinstance(o, bytes):
        return None

    return [o[i] for i in range(4, 44, 4)]


def set_fan_curve(arr: Sequence[int], lim: Sequence[int] | None = None):
    logger.info(f"Setting fan curve to:\n{arr}")
    if len(arr) != 10:
        logger.error(f"Invalid fan curve length: {len(arr)}. Should be 10.")
        return False
    if any(not isinstance(d, int) for d in arr):
        logger.error(f"Curve has null value, not setting.")
        return False

    if lim:
        for a, b in zip(arr, lim):
            if a < b:
                logger.error(
                    f"Not set. Fan curve does not comply with limit:\n{len(arr)}"
                )
                return False

    return call(
        r"\_SB.GZFD.WMAB",
        [
            0,
            0x06,
            bytes(
                [
                    0x00,
                    0x00,
                    0x0A,
                    0x00,
                    0x00,
                    0x00,
                    arr[0],
                    0x00,
                    arr[1],
                    0x00,
                    arr[2],
                    0x00,
                    arr[3],
                    0x00,
                    arr[4],
                    0x00,
                    arr[5],
                    0x00,
                    arr[6],
                    0x00,
                    arr[7],
                    0x00,
                    arr[8],
                    0x00,
                    arr[9],
                    0x00,
                    0x00,
                    0x0A,
                    0x00,
                    0x00,
                    0x00,
                    0x0A,
                    0x00,
                    0x14,
                    0x00,
                    0x1E,
                    0x00,
                    0x28,
                    0x00,
                    0x32,
                    0x00,
                    0x3C,
                    0x00,
                    0x46,
                    0x00,
                    0x50,
                    0x00,
                    0x5A,
                    0x00,
                    0x64,
                    0x00,
                    0x00,
                ]
            ),
        ],
    )


def set_power_light_v1(enabled: bool):
    logger.debug(f"Setting power light status.")
    return call(r"\_SB.GZFD.WMAF", [0, 0x02, bytes([0x03, int(enabled), 0x00])])


def get_power_light_v1():
    logger.debug(f"Getting power light status.")
    if not call(r"\_SB.GZFD.WMAF", [0, 0x01, 0x03], risky=False):
        return None
    o = read()
    if isinstance(o, bytes) and len(o) == 2:
        return bool(o[0])
    return None


def set_power_light(enabled: bool, suspend: bool = False):
    logger.debug(f"Setting power light status.")
    if enabled:
        if suspend:
            cb = 0x03
        else:
            cb = 0x02
    else:
        cb = 0x01
    return call(
        r"\_SB.GZFD.WMAF",
        [0, 0x02, bytes([0x024 if suspend else 0x04, 0x00, cb])],
    )


def get_power_light(suspend: bool = False):
    logger.debug(f"Getting power light status.")
    if not call(r"\_SB.GZFD.WMAF", [0, 0x01, 0x024 if suspend else 0x04], risky=False):
        return None
    o = read()
    if isinstance(o, bytes) and len(o) == 2:
        return o[1] == (0x03 if suspend else 0x02)
    return None


def get_bios_version():
    raw = None
    try:
        with open("/sys/class/dmi/id/bios_version") as f:
            raw = f.read()
        return int(raw.replace("N3CN", "").split("WW")[0].strip())
    except Exception as e:
        logger.error(f"Failed to get BIOS version from '{raw}' with error:\n{e}")
        return 1


def get_feature(id: int):
    if not call(
        r"\_SB.GZFD.WMAE",
        [0, 0x11, int.to_bytes(id, length=4, byteorder="little", signed=False)],
        risky=False,
    ):
        return None

    return read()


def set_feature(id: int, value: int):
    return call(
        r"\_SB.GZFD.WMAE",
        [
            0,
            0x12,
            int.to_bytes(id, length=4, byteorder="little", signed=False)
            + int.to_bytes(value, length=4, byteorder="little", signed=False),
        ],
    )


def set_tdp_mode(mode: TdpMode):
    logger.info(f"Setting tdp mode to '{mode}'.")
    match mode:
        case "quiet":
            b = 0x01
        case "balanced":
            b = 0x02
        case "performance":
            b = 0x03
        case "custom":
            b = 0xFF
        case _:
            logger.error(f"TDP mode '{mode}' is unknown. Not setting.")
            return False

    return call(r"\_SB.GZFD.WMAA", [0, 0x2C, b])


def get_tdp_mode() -> TdpMode | None:
    logger.debug(f"Retrieving TDP Mode.")
    if not call(r"\_SB.GZFD.WMAA", [0, 0x2D, 0], risky=False):
        logger.error(f"Failed retrieving TDP Mode.")
        return None

    match read():
        case 0x01:
            return "quiet"
        case 0x02:
            return "balanced"
        case 0x03:
            return "performance"
        case 0xFF:
            return "custom"
        case v:
            logger.error(f"TDP mode '{v}' is unknown")
            return None


def get_steady_tdp():
    logger.debug(f"Retrieving steady TDP.")
    return get_feature(0x0102FF00)


def get_fast_tdp():
    logger.debug(f"Retrieving fast TDP.")
    return get_feature(0x0103FF00)


def get_slow_tdp():
    logger.debug(f"Retrieving slow TDP.")
    return get_feature(0x0101FF00)


def get_charge_limit():
    logger.debug(f"Retrieving charge limit.")
    return get_feature(0x03010001)


def set_charge_limit(enable: bool):
    logger.info(f"Setting charge limit (80 %) to {enable}.")
    return set_feature(0x03010001, enable)


def set_steady_tdp(val: int):
    logger.info(f"Setting steady TDP to {val}.")
    return set_feature(0x0102FF00, val)


def set_fast_tdp(val: int):
    logger.info(f"Setting fast TDP to {val}.")
    return set_feature(0x0103FF00, val)


def set_slow_tdp(val: int):
    logger.info(f"Setting slow TDP to {val}.")
    return set_feature(0x0101FF00, val)


def get_full_fan_speed():
    logger.debug(f"Getting full fan speed.")
    return get_feature(0x04020000)


def set_full_fan_speed(enable: bool):
    logger.info(f"Setting full fan mode to {enable}.")
    return set_feature(0x04020000, int(enable))



import logging
import time
from typing import Sequence, cast

from hhd.plugins import Context, Event, HHDPlugin, load_relative_yaml
from hhd.plugins.conf import Config

from adjustor.core.lenovo import (
    MIN_CURVE,
    TdpMode,
    get_bios_version,
    get_full_fan_speed,
    get_power_light,
    get_power_light_v1,
    get_tdp_mode,
    set_charge_limit,
    set_fan_curve,
    set_fast_tdp,
    set_full_fan_speed,
    set_power_light,
    set_power_light_v1,
    set_slow_tdp,
    set_steady_tdp,
    set_tdp_mode,
)
from adjustor.i18n import _

logger = logging.getLogger(__name__)

APPLY_DELAY = 0.5
TDP_DELAY = 0


class LenovoDriverPlugin(HHDPlugin):
    def __init__(self, legion_s=False) -> None:
        self.name = f"adjustor_lenovo"
        self.priority = 6
        self.log = "adjl"
        self.enabled = False
        self.initialized = False
        self.enforce_limits = True
        self.startup = True
        self.old_conf = None
        self.old_mode = None
        self.prev_charge_limit = None
        self.sys_tdp = False
        self.fan_curve_set = False
        self.legion_s = legion_s
        if legion_s:
            # AC 40/43/53W
            # DC 33/33/43W
            self.max_watts = 40
            self.max_watts_sppt = 43
            self.max_watts_fppt = 53
            self.fppt_ratio = 53 / 40
            self.performance_tdp = 25
        else:
            self.max_watts = 30
            self.max_watts_sppt = 32
            self.max_watts_fppt = 41
            self.fppt_ratio = 41/30
            self.performance_tdp = 20

        if legion_s:
            self.power_light_v2 = False
        else:
            bios_version = get_bios_version()
            logger.info(f"Lenovo BIOS version: {bios_version}")
            self.power_light_v2 = bios_version >= 35

        self.queue_fan = None
        self.queue_tdp = None
        self.new_tdp = None
        self.new_mode = None
        self.old_target = None

    def settings(self):
        if not self.enabled:
            self.initialized = False
            self.old_conf = None
            self.startup = True
            return {}

        self.initialized = True
        out = {"tdp": {"lenovo": load_relative_yaml("settings.yml")}}
        if self.legion_s:
            out["tdp"]["lenovo"]["children"]["power_light"]["title"] = _("Power Light")
        if not self.power_light_v2:
            del out["tdp"]["lenovo"]["children"]["power_light_sleep"]
        if not self.enforce_limits:
            out["tdp"]["lenovo"]["children"]["tdp"]["modes"]["custom"]["children"][
                "tdp"
            ]["max"] = 40
        else:
            out["tdp"]["lenovo"]["children"]["tdp"]["modes"]["custom"]["children"][
                "tdp"
            ]["max"] = self.max_watts
        if self.legion_s:
            out["tdp"]["lenovo"]["children"]["tdp"]["modes"]["performance"][
                "unit"
            ] = f"25W"
            out["tdp"]["lenovo"]["children"]["tdp"]["modes"]["custom"][
                "unit"
            ] = f"→ 33/40W"
        return out

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit

    def update(self, conf: Config):
        self.enabled = conf["hhd.settings.tdp_ready"].to(bool)
        new_enforce_limits = conf["hhd.settings.enforce_limits"].to(bool)
        new_lims = new_enforce_limits != self.enforce_limits
        self.enforce_limits = new_enforce_limits

        if not self.enabled or not self.initialized or new_lims:
            self.old_conf = None
            self.startup = True
            self.fan_curve_set = False
            return

        #
        # Checks
        #

        # Initialize values so we do not query them all the time
        tdp_reset = self.startup
        if self.startup:
            conf["tdp.lenovo.ffss"] = get_full_fan_speed()
            if self.power_light_v2:
                conf["tdp.lenovo.power_light"] = get_power_light(suspend=False)
                conf["tdp.lenovo.power_light_sleep"] = get_power_light(suspend=True)
            else:
                conf["tdp.lenovo.power_light"] = get_power_light_v1()

        # If not old config, exit, as values can not be set
        if not self.old_conf:
            self.old_conf = conf["tdp.lenovo"]
            return

        curr = time.time()

        #
        # Other options
        #
        ffss = conf["tdp.lenovo.ffss"].to(bool)
        if ffss is not None and ffss != self.old_conf["ffss"].to(bool):
            set_full_fan_speed(ffss)

        power_light = conf["tdp.lenovo.power_light"].to(bool)
        if power_light is not None and power_light != self.old_conf["power_light"].to(
            bool
        ):
            if self.power_light_v2:
                set_power_light(power_light, suspend=False)
            else:
                set_power_light_v1(power_light)
        if self.power_light_v2:
            power_light_sleep = conf["tdp.lenovo.power_light_sleep"].to(bool)
            if (
                power_light_sleep != self.old_conf["power_light_sleep"].to(bool)
                and power_light_sleep is not None
            ):
                set_power_light(power_light_sleep, suspend=True)

        charge_limit = conf.get("tdp.lenovo.charge_limit", False)
        if charge_limit != self.prev_charge_limit:
            set_charge_limit(charge_limit)
            self.prev_charge_limit = charge_limit

        #
        # TDP
        #

        # Update tdp mode if user changed through the app
        new_target = None
        new_tdp = self.new_tdp
        self.new_tdp = None
        new_mode = self.new_mode
        notify_tdp = False
        self.new_mode = None
        if new_tdp:
            # For TDP values received from steam, set the appropriate
            # mode to get a better experience.
            if new_tdp == 8:
                mode = "quiet"
            elif new_tdp == 15:
                mode = "balanced"
            elif new_tdp == self.performance_tdp:
                mode = "performance"
            else:
                mode = "custom"
            conf["tdp.lenovo.tdp.mode"] = mode
        elif new_mode:
            mode = new_mode
            conf["tdp.lenovo.tdp.mode"] = mode
        else:
            mode = conf["tdp.lenovo.tdp.mode"].to(str)
        if mode is not None and mode != self.old_mode:
            self.old_mode = mode
            set_tdp_mode(cast(TdpMode, mode))
            tdp_reset = True

        # Grab from power button
        new_mode = get_tdp_mode()
        if new_mode != mode:
            if not new_tdp:
                self.sys_tdp = False
            tdp_reset = True
            if self.old_mode is not None:
                notify_tdp = True
            self.old_mode = mode
        conf["tdp.lenovo.tdp.mode"] = new_mode

        # Reset fan curve on mode change
        # Has to happen before setting the stdp, ftdp values, in case
        # we are in custom mode
        fan_mode = conf["tdp.lenovo.fan.mode"].to(str)
        if fan_mode != self.old_conf["fan.mode"].to(str) and fan_mode != "manual":
            tdp_mode = get_tdp_mode()
            if tdp_mode:
                set_tdp_mode("performance")
                time.sleep(TDP_DELAY)
                set_tdp_mode(tdp_mode)
                tdp_reset = True

        # Handle EPP for presets
        if tdp_reset and new_mode != "custom":
            match new_mode:
                case "quiet":
                    new_target = "power"
                case "balanced":
                    new_target = "balanced"
                case "performance":
                    new_target = "performance"

        # In custom mode, re-apply settings with debounce
        if new_mode == "custom":
            # Check user changed values
            steady = conf["tdp.lenovo.tdp.custom.tdp"].to(int)
            if new_tdp:
                steady = new_tdp
                conf["tdp.lenovo.tdp.custom.tdp"] = steady

            old_steady = steady
            if self.enforce_limits:
                steady = min(max(steady, 4), self.max_watts)
            else:
                steady = min(max(steady, 0), 50)
            if old_steady != steady:
                conf["tdp.lenovo.tdp.custom.tdp"] = steady

            steady_updated = steady and steady != self.old_conf["tdp.custom.tdp"].to(
                int
            )
            if steady_updated and not new_tdp:
                self.sys_tdp = False

            if self.startup and (steady > self.max_watts or steady < 7):
                logger.warning(
                    f"TDP ({steady}) outside the device spec. Resetting for stability reasons."
                )
                steady = self.max_watts
                conf["tdp.lenovo.tdp.custom.tdp"] = self.max_watts
                steady_updated = True

            boost = conf["tdp.lenovo.tdp.custom.boost"].to(bool)
            boost_updated = boost != self.old_conf["tdp.custom.boost"].to(bool)

            # If yes, queue an update
            # Debounce
            if steady_updated or boost_updated or tdp_reset:
                self.queue_tdp = curr + APPLY_DELAY

            if self.queue_tdp and self.queue_tdp < curr:
                self.queue_tdp = None
                if boost:
                    set_steady_tdp(steady)
                    time.sleep(TDP_DELAY)
                    set_slow_tdp(steady + (self.max_watts_sppt - self.max_watts))
                    time.sleep(TDP_DELAY)
                    set_fast_tdp(
                        min(
                            self.max_watts_fppt,
                            int(steady * self.fppt_ratio),
                        )
                    )
                else:
                    set_steady_tdp(steady)
                    time.sleep(TDP_DELAY)
                    set_slow_tdp(steady)
                    time.sleep(TDP_DELAY)
                    set_fast_tdp(steady)

                # Handle EPP for custom mode
                if steady < 12:
                    new_target = "power"
                elif steady <= 20:
                    new_target = "balanced"
                else:
                    new_target = "performance"

        # Handle EPP application
        if new_target and new_target != self.old_target:
            self.old_target = new_target
            self.emit({"type": "energy", "status": new_target})

        # Fan curve stuff
        # If tdp reset, so was the curve
        if tdp_reset:
            # 2x to apply after tdp
            self.queue_fan = curr + 2 * APPLY_DELAY

        # Handle fan curve resets
        if conf["tdp.lenovo.fan.manual.reset"].to(bool):
            conf["tdp.lenovo.fan.manual.reset"] = False
            for i, v in enumerate(MIN_CURVE):
                conf[f"tdp.lenovo.fan.manual.st{(i + 1)*10}"] = v

        # Handle fan curve limits
        for i, v in enumerate(MIN_CURVE):
            val = conf[f"tdp.lenovo.fan.manual.st{(i + 1)*10}"].to(int)
            old_val = val
            if conf["tdp.lenovo.fan.manual.enforce_limits"].to(bool) and val < v:
                val = v

            val = max(min(val, 115), 0)
            if old_val != val:
                conf[f"tdp.lenovo.fan.manual.st{(i + 1)*10}"] = val

        # Check if fan curve has changed
        # Use debounce logic on these changes
        for i in (10, 20, 30, 40, 50, 60, 70, 80, 90, 100):
            if conf[f"tdp.lenovo.fan.manual.st{i}"].to(int) != self.old_conf[
                f"fan.manual.st{i}"
            ].to(int):
                self.queue_fan = curr + APPLY_DELAY

        apply_curve = (
            self.queue_fan and self.queue_fan < curr
        ) or not self.fan_curve_set
        if conf["tdp.lenovo.fan.mode"].to(str) == "manual" and apply_curve:
            try:
                set_fan_curve(
                    [
                        conf[f"tdp.lenovo.fan.manual.st{i}"].to(int)
                        for i in (10, 20, 30, 40, 50, 60, 70, 80, 90, 100)
                    ]
                )
            except Exception as e:
                logger.error(f"Could not set fan curve. Error:\n{e}")
            self.fan_curve_set = True
            self.queue_fan = None

        # Show steam message
        if self.sys_tdp:
            conf["tdp.lenovo.cycle_info"] = _("Steam is controlling TDP")
        else:
            conf["tdp.lenovo.cycle_info"] = _("Legion Button + Y changes TDP Mode")

        # Save current config
        self.old_conf = conf["tdp.lenovo"]

        if notify_tdp:
            notify_tdp = False
            if conf.get("tdp.lenovo.tdp_rgb", False):
                self.emit({"type": "special", "event": f"tdp_cycle_{new_mode}"})  # type: ignore

        if self.startup:
            self.startup = False

    def notify(self, events: Sequence[Event]):
        for ev in events:
            if ev["type"] == "tdp":
                self.new_tdp = ev["tdp"]
                self.sys_tdp = ev["tdp"] is not None
            if ev["type"] == "ppd":
                match ev["status"]:
                    case "power":
                        self.new_mode = "quiet"
                    case "balanced":
                        self.new_mode = "balanced"
                    case "performance":
                        self.new_mode = "performance"

    def close(self):
        pass



import logging
import os
import subprocess
from typing import NamedTuple, Sequence

logger = logging.getLogger(__name__)


class Command(NamedTuple):
    method: str
    args: Sequence[bytes]


def initialize():
    try:
        o = subprocess.run(["modprobe", "acpi_call"], capture_output=True)
        logger.info(f"'acpi_call' modprobe output:\n{(o.stdout + o.stderr).decode()}".strip())
        return True
    except Exception as e:
        logger.warning(f"Failed loading acpi_call with error:\n{e}")
        return False


def check_perms():
    try:
        with open("/proc/acpi/call", "wb") as f:
            return f.writable()
    except Exception as e:
        logger.error(f"Could open acpi_call file ('/proc/acpi/call'). Error:\n{e}")
        return False


def call(method: str, args: Sequence[bytes | int], risky: bool = True):
    cmd = method
    for arg in args:
        if isinstance(arg, int):
            cmd += f" 0x{arg:02x}"
        else:
            cmd += f" b{arg.hex()}"

    log = logger.info if risky else logger.debug
    log(f"Executing ACPI call:\n'{cmd}'")

    try:
        with open("/proc/acpi/call", "wb") as f:
            f.write(cmd.encode())
        return True
    except Exception as e:
        logger.error(f"ACPI Call failed with error:\n{e}")
        return False


def read():
    with open("/proc/acpi/call", "rb") as f:
        d = f.read().decode().strip()

    if d == "not called\0":
        return None
    if d.startswith("0x") and d.endswith("\0"):
        return int(d[:-1], 16)
    if d.startswith("{") and d.endswith("}\0"):
        bs = d[1:-2].split(", ")
        return bytes(int(b, 16) for b in bs)
    assert False, f"Return value '{d}' supported yet or was truncated."



import logging
import time
from threading import Event as TEvent, Lock, Thread
from typing import Sequence
import os

from hhd.plugins import Context, Event, HHDPlugin, load_relative_yaml
from hhd.plugins.conf import Config

from adjustor.core.alib import AlibParams, DeviceParams, alib
from adjustor.core.fan import fan_worker, get_fan_info
from adjustor.core.platform import get_platform_choices, set_platform_profile
from adjustor.i18n import _

logger = logging.getLogger(__name__)

APPLY_DELAY = 0.7


def set_charge_limit(bat: str, lim: int):
    try:
        logger.info(f"Setting charge limit to {lim:d} %.")
        with open(bat, "w") as f:
            f.write(f"{lim}\n")
        return True
    except Exception as e:
        logger.error(f"Failed to write battery limit with error:\n{e}")
        return False


def set_charge_bypass_type(bat: str, type: str):
    match type:
        case "disabled":
            val = "Standard"
        case "awake":
            val = "BypassS0"
        case "always":
            val = "Bypass"
        case _:
            logger.error(f"Invalid charge bypass type: {type}")
            return False

    try:
        logger.info(f"Setting charge type to '{val}' (for bypass '{type}').")
        with open(bat, "w") as f:
            f.write(f"{val}\n")
        return True
    except Exception as e:
        logger.error(f"Failed to write battery bypass with error:\n{e}")
        return False


def set_charge_bypass_behaviour(bat: str, type: str):
    match type:
        case "disabled":
            val = "auto"
        case "awake":
            val = "inhibit-charge-awake"
        case "always":
            val = "inhibit-charge"
        case _:
            logger.error(f"Invalid charge bypass type: {type}")
            return False

    try:
        logger.info(f"Setting charge type to '{val}' (for bypass '{type}').")
        with open(bat, "w") as f:
            f.write(f"{val}\n")
        return True
    except Exception as e:
        logger.error(f"Failed to write battery bypass with error:\n{e}")
        return False


def set_charge_bypass(bat: str, type: str):
    if "charge_type" in bat:
        return set_charge_bypass_type(bat, type)
    elif "charge_behaviour" in bat:
        return set_charge_bypass_behaviour(bat, type)
    else:
        logger.error(f"Unknown charge bypass file: {bat}")
        return False


class BatteryPlugin(HHDPlugin):

    def __init__(self, always_enable: bool = False) -> None:
        self.name = f"adjustor_battery"
        self.priority = 9
        self.log = "batt"
        self.enabled = False
        self.initialized = False
        self.startup = False

        self.queue_charge_limit = None
        self.charge_bypass_fn = None
        self.bypass_awake = True
        self.charge_limit_fn = None
        self.charge_bypass_prev = None
        self.charge_limit_prev = None

        self.always_enable = always_enable

    def settings(self):
        if not self.enabled:
            self.initialized = False
            return {}

        self.initialized = True
        out = {"tdp": {"battery": load_relative_yaml("battery.yml")}}

        if not self.charge_limit_fn:
            del out["tdp"]["battery"]["children"]["charge_limit"]
        if not self.charge_bypass_fn:
            del out["tdp"]["battery"]["children"]["charge_bypass"]
        elif not self.bypass_awake:
            del out["tdp"]["battery"]["children"]["charge_bypass"]["options"]["awake"]

        return out

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit
        self.startup = True

        for bat in os.listdir("/sys/class/power_supply"):
            if not bat.startswith("BAT"):
                continue

            with open(f"/sys/class/power_supply/{bat}/type") as f:
                if "Battery" not in f.read():
                    continue

            base = f"/sys/class/power_supply/{bat}"
            if os.path.exists(f"{base}/charge_control_end_threshold"):
                self.charge_limit_fn = f"{base}/charge_control_end_threshold"
            if os.path.exists(f"{base}/charge_type"):
                try:
                    with open("/sys/devices/virtual/dmi/id/sys_vendor") as f:
                        supports = "ONE-NETBOOK" in f.read()
                except Exception:
                    supports = False

                if supports:
                    self.charge_bypass_fn = f"{base}/charge_type"
                    self.bypass_awake = True
                else:
                    logger.warning(
                        "Found charge type, but charge bypass is only supported on OneXPlayer currently."
                    )
                    self.charge_bypass_fn = None
                    self.bypass_awake = False
            if os.path.exists(f"{base}/charge_behaviour"):
                self.charge_bypass_fn = f"{base}/charge_behaviour"
                try:
                    with open(self.charge_bypass_fn) as f:
                        self.bypass_awake = "inhibit-charge-awake" in f.read()
                except Exception:
                    logger.error(
                        "Failed to read charge behaviour file, assuming it is not supported."
                    )
                    self.charge_bypass_fn = None
                    self.bypass_awake = False
            if self.charge_bypass_fn or self.charge_limit_fn:
                logger.info(
                    f"Found battery '{bat}' with:\nBattery Bypass:\n{self.charge_bypass_fn}\nBattery Limit:\n{self.charge_limit_fn}."
                )
                break

    def update(self, conf: Config):
        self.enabled = self.always_enable or conf.get("hhd.settings.tdp_ready", False)

        if not self.initialized:
            return

        curr = time.time()

        if self.charge_bypass_fn:
            bypass = conf["tdp.battery.charge_bypass"].to(str)
            if self.charge_bypass_prev != bypass:
                self.charge_bypass_prev = bypass

                if bypass != "disabled" or not self.startup:
                    set_charge_bypass(self.charge_bypass_fn, bypass)

        # Charge limit
        if self.charge_limit_fn:
            lim = conf["tdp.battery.charge_limit"].to(str)
            if lim != self.charge_limit_prev:
                self.queue_charge_limit = curr + APPLY_DELAY
                self.charge_limit_prev = lim

            if self.startup or (
                self.queue_charge_limit and self.queue_charge_limit < curr
            ):
                self.queue_charge_limit = None
                self.charge_limit_prev = lim

                match lim:
                    case "p65":
                        set_charge_limit(self.charge_limit_fn, 65)
                    case "p70":
                        set_charge_limit(self.charge_limit_fn, 70)
                    case "p80":
                        set_charge_limit(self.charge_limit_fn, 80)
                    case "p85":
                        set_charge_limit(self.charge_limit_fn, 85)
                    case "p90":
                        set_charge_limit(self.charge_limit_fn, 90)
                    case "p95":
                        set_charge_limit(self.charge_limit_fn, 95)
                    case "disabled":
                        # Avoid writing charge limit on startup if
                        # disabled, so that if user does not use us
                        # we do not overwrite their setting.
                        if not self.startup:
                            set_charge_limit(self.charge_limit_fn, 100)

        self.startup = False



# The objective of this fan algorithm is simple: receive a fan curve for a specific
# power profile, the current system temperature, and the current fan speed.
# Then, have a single variable memory: the current acceleration.
# At each point, use a derived jerk to modify the acceleration and move the
# speed to the desired point smoothly.

# Maximum typical transition span of the fan speed (e.g., from 20% to 70%)
# This transition span should take exactly the seconds specified below
# E.g., if the fan begins to move from 20% to 70% with 0 acceleration, where
# the temperature is at 70C, it should take at most 5 seconds.
SPEED_SPAN = 0.5

# The ratio of the curve to begin decelerating the fan change rpm/s^2
# E.g., when going from 20% to 70%, the fan should begin decelerating
# at 0.7*(70-20) + 20 = 62% speed.
DECEL_RATIO = 0.5

# The temperature at which the fan should use the high accel speed.
HIGH_TEMP_EDGE = 65
HIGH_TEMP_JUNCTION = 75

# Allow for a gradual transition if the temperature is low
# Speed up if not.
ACCEL_UP_LOWT_T = 30
ACCEL_UP_HIGH_T = 20
# Penalize going down to avoid dithering
ACCEL_DOWN_T = 40

JERK_TOLERANCE = 0.9
MAX_ACCEL = 0.4
SETPOINT_DEVIATION = 0.01

UPDATE_FREQUENCY = 5
UPDATE_T = 1 / UPDATE_FREQUENCY
SETPOINT_UPDATE_FREQUENCY = 1
SETPOINT_UPDATE_T = 1 / SETPOINT_UPDATE_FREQUENCY


def _calculate_jerk(speed_span, decel_ratio, freq, time):
    """Calculate the required positive and negative jerks such that the
    fan can cross the speed span (e.g., 50%) in the specified time (e.g., 5s)
    given an update frequency (e.g., 3), and the point at which it should start
    decelerating (e.g., 0.3).
    """

    jerk_accel = (2 * speed_span) / ((time * freq) ** 2) / ((1 - decel_ratio) ** 2)
    jerk_decel = -(1 - decel_ratio) / decel_ratio * jerk_accel
    return jerk_accel, jerk_decel


def calculate_jerk(t_target: float, increase: bool, junction: bool):
    """Calculate the jerk based on the target temperature and whether the fan
    speed should increase or decrease to reach it.

    Allow for specifying whether the temperature probe is in the junction or the
    edge, as junction reaches thermal saturation faster and at higher temperatures."""
    if not increase:
        return _calculate_jerk(SPEED_SPAN, DECEL_RATIO, UPDATE_FREQUENCY, ACCEL_DOWN_T)

    if (junction and t_target > HIGH_TEMP_JUNCTION) or (
        not junction and t_target > HIGH_TEMP_EDGE
    ):
        return _calculate_jerk(
            SPEED_SPAN, DECEL_RATIO, UPDATE_FREQUENCY, ACCEL_UP_HIGH_T
        )

    return _calculate_jerk(SPEED_SPAN, DECEL_RATIO, UPDATE_FREQUENCY, ACCEL_UP_LOWT_T)


def move_to_setpoint(v_curr, a_curr, jerk_accel, jerk_decel, v_target):
    """Update the current fan speed and acceleration by either using jerk_accel
    which will increase acceleration to meet the target speed or jerk_decel
    which will begin to decrease it to 0.

    The choice between jerk_accel and jerk_decel is made by calculating the
    minimum negative jerk required to decelerate to a=0 when reaching the target.
    If the minimum jerk is smaller than jerk_decel, we use jerk_accel.
    To avoid overshoots, a tolerance is used to start decelerating a bit earlier
    than when the minimum jerk reaches the value of jerk_decel.
    """

    # Flip the jerks if the target is lower than the current speed
    diff = v_target - v_curr
    if diff < 0:
        jerk_accel = -jerk_accel
        jerk_decel = -jerk_decel

    correct_direction = (diff > 0 and a_curr > 0) or (diff < 0 and a_curr < 0)
    non_zero = abs(diff) > 1e-3

    # Always accelerate if we are on the right direction or speed is zero
    # Start decelerating once we run out of opposite jerk.
    accel = True
    if correct_direction and non_zero:
        min_jerk_neg = -(a_curr**2) / 2 / diff
        accel = abs(min_jerk_neg) < JERK_TOLERANCE * abs(jerk_decel)

    if accel:
        jerk = jerk_accel
    else:
        jerk = jerk_decel

    # Calculate the new acceleration
    a_new = a_curr + jerk
    v_new = v_curr + a_new
    return v_new, a_new


def sanitize_fan_values(v: float, a: float):
    return max(0, min(1, v)), max(-MAX_ACCEL, min(MAX_ACCEL, a))


def has_reached_setpoint(v_curr, a_curr, v_target):
    """Check if the current fan speed has reached the target speed.

    If true, lower the update rate and set a_curr to 0 to avoid dithering.
    """
    return abs(v_curr - v_target) < SETPOINT_DEVIATION


def update_setpoint(temp: float, curr: int, fan_curve: dict[int, float]):
    """Update the setpoint given the current temperature, fan curve, and previous setpoint.

    Fan curve is a dictionary of increasing temperatures to fan speeds.
    """

    targets = list(fan_curve.keys())
    assert curr in targets, "Current setpoint not in fan curve"

    idx = targets.index(curr)

    # Add some hysterisis to avoid dithering
    if idx > 0:
        prev = targets[idx - 1]
        if temp < prev:
            return prev

    if idx < len(targets) - 1:
        next = targets[idx + 1]
        if temp > next:
            return next

    return curr


def get_initial_setpoint(temp: float, fan_curve: dict[int, float]):
    """Get the initial setpoint given the current temperature and fan curve.

    Fan curve is a dictionary of increasing temperatures to fan speeds.
    """

    targets = list(fan_curve.keys())
    for idx, target in enumerate(targets):
        if temp < target:
            return targets[idx - 1] if idx else targets[0]
    return targets[-1]



import argparse
import fcntl
import logging
import os
import signal
import subprocess
import sys
import time
from os.path import join
from threading import Condition
from threading import Event as TEvent
from threading import RLock
from time import sleep
from typing import Callable, Sequence

import pkg_resources

from .logging import set_log_plugin, setup_logger, update_log_plugins
from .plugins import (
    Config,
    Emitter,
    Event,
    HHDAutodetect,
    HHDPlugin,
    HHDSettings,
    load_relative_yaml,
)
from .plugins.settings import (
    Validator,
    get_default_state,
    get_settings_hash,
    load_blacklist_yaml,
    load_profile_yaml,
    load_state_yaml,
    merge_settings,
    parse_defaults,
    save_blacklist_yaml,
    save_profile_yaml,
    save_state_yaml,
    validate_config,
)
from .utils import (
    GIT_HHD,
    HHD_DEV_DIR,
    expanduser,
    get_ac_status,
    get_ac_status_fn,
    get_context,
    get_os,
    refresh_is_steam_running,
    switch_priviledge,
)

logger = logging.getLogger(__name__)

DEBUG_MODE = bool(os.environ.get("HHD_DEBUG", False))
CONFIG_DIR = os.environ.get("HHD_CONFIG_DIR", "/etc/hhd")
TOKEN_FILE = os.environ.get("HHD_TOKEN_FILE", "/tmp/hhd/token")
PLUGIN_WHITELIST = os.environ.get("HHD_PLUGINS", "")

ERROR_DELAY = 5
INIT_DELAY = 0.4
POLL_DELAY = 2
SLEEP_MIN_T = 8


class EmitHolder(Emitter):
    def __init__(self, condition: Condition, ctx, info) -> None:
        self._events = []
        self._condition = condition
        super().__init__(ctx=ctx, info=info)

    def __call__(self, event: Event | Sequence[Event]) -> None:
        with self._condition:
            if isinstance(event, Sequence):
                self._events.extend(event)
            else:
                self._events.append(event)
            self._condition.notify_all()

    def get_events(self, timeout: int = -1) -> Sequence[Event]:
        with self._condition:
            if not self._events and timeout != -1:
                self._condition.wait()
            ev = self._events
            self._events = []
            return ev

    def has_events(self):
        with self._condition:
            return bool(self._events)


def notifier(ev: TEvent, cond: Condition):
    def _inner(sig, frame):
        with cond:
            ev.set()
            cond.notify_all()

    return _inner


def get_wakeup_count():
    try:
        with open("/sys/power/wakeup_count", "r") as f:
            return int(f.read().strip())
    except Exception:
        return -1


def print_token(ctx):
    try:
        with open(TOKEN_FILE, "r") as f:
            token = f.read().strip()

        logger.info(f'Current HHD token (for user "{ctx.name}") is: "{token}"')
    except Exception as e:
        logger.error(f"Token not found or could not be read, error:\n{e}")


def main():
    parser = argparse.ArgumentParser(
        prog="HHD: Handheld Daemon main interface.",
        description="Handheld Daemon is a daemon for managing the quirks inherent in handheld devices.",
    )
    parser.add_argument(
        "-u",
        "--user",
        default=None,
        help="This argument has been deprecated. Handheld Daemon will always use /etc/hhd moving forward.",
        dest="user",
    )
    parser.add_argument(
        "command",
        nargs="*",
        default=[],
        help="The command to run. If empty, run as daemon. Right now, only the command `token` is supported.",
    )
    args = parser.parse_args()
    user = args.user

    # Setup temporary logger for permission retrieval
    if user:
        ctx = get_context(user)
    else:
        ctx = None

    detectors: dict[str, HHDAutodetect] = {}
    plugins: dict[str, Sequence[HHDPlugin]] = {}
    cfg_fds = []
    switch_root = None

    # HTTP data
    https = None
    prev_http_cfg = None
    updated = False
    last_event = None
    ac_fn = get_ac_status_fn()
    info = Config()
    info["ac"] = None
    ac_status = None

    # Check we are in a virtual environment
    # TODO: Improve
    exe_python = sys.executable

    try:
        # Create nested hhd dir
        # This might mess up permissions in upward directories
        # So try to deescalate
        hhd_dir = CONFIG_DIR
        try:
            os.makedirs(hhd_dir, exist_ok=True)
        except Exception:
            pass

        set_log_plugin("main")
        setup_logger(join(CONFIG_DIR, "log"), ctx=ctx)

        if args.command:
            if args.command[0] == "token":
                print_token(ctx)
                return
            else:
                logger.error(f"Command '{args.command[0]}' is unknown. Ignoring...")

        # Run a one time migration
        # TODO: remove this e.g., after a year
        try:
            migrate_fn = join(hhd_dir, ".migrated")
            state_fn = join(CONFIG_DIR, "state.yml")
            if not os.path.isfile(migrate_fn) and not os.path.isfile(state_fn):
                # Always touch migration file so we only run on a fresh install
                with open(migrate_fn, "w") as f:
                    f.write("")

                for user in os.listdir("/home"):
                    old_state = expanduser("~/.config/hhd/state.yml", ctx)
                    if os.path.isfile(old_state):
                        import shutil

                        shutil.copy(old_state, state_fn)
                        os.chmod(state_fn, 0o644)
                        os.chown(state_fn, 0, 0)
                        logger.warning(
                            f"Migrated state file {old_state} to {state_fn}."
                        )
                        break
        except Exception as e:
            logger.error(f"Could not migrate state file. Error:\n{e}")

        # Lock a file to ensure hhd is only running once, or wait to unlock
        # THis is for legacy enabled services with --user
        lock_fn = join(hhd_dir, "lock")
        lock_fd = -1
        if os.environ.get("HHD_SWITCH_ROOT", None) != "1":
            try:
                logger.info("Trying to acquire hhd lock...")
                lock_fd = os.open(lock_fn, os.O_RDWR | os.O_CREAT)
                fcntl.flock(lock_fd, fcntl.LOCK_EX)
            except Exception as e:
                logger.error(f"Could not acquire lock for hhd.\nError:\n{e}")

        # Get OS Info
        info["os"] = get_os()

        # Use blacklist
        blacklist_fn = join(hhd_dir, "plugins.yml")
        blacklist = load_blacklist_yaml(blacklist_fn)

        logger.info(f"Running autodetection...")

        detector_names = []
        whitelist = PLUGIN_WHITELIST.split(",") if PLUGIN_WHITELIST else []
        for autodetect in pkg_resources.iter_entry_points("hhd.plugins"):
            name = autodetect.name
            detector_names.append(name)
            if name in blacklist:
                logger.info(f"Skipping blacklisted provider '{name}'.")
            if whitelist and name not in whitelist:
                logger.info(f"Skipping provider '{name}' due to whitelist.")
                continue

            detectors[autodetect.name] = autodetect.resolve()

        # Save new blacklist file
        save_blacklist_yaml(blacklist_fn, detector_names, blacklist)

        logger.info(f"Found plugin providers: {', '.join(list(detectors))}")

        for name, autodetect in detectors.items():
            try:
                plugins[name] = autodetect([])
            except KeyboardInterrupt:
                raise
            except Exception as e:
                if DEBUG_MODE:
                    raise e
                logger.error(f"Could not load plugins from provider '{name}'.")
                try:
                    from rich import get_console

                    get_console().print_exception()
                except Exception:
                    import traceback

                    traceback.print_exc()

        plugin_str = "Loaded the following plugins:"
        for pkg_name, sub_plugins in plugins.items():
            if not sub_plugins:
                continue
            plugin_str += (
                f"\n  - {pkg_name:>8s}: {', '.join(p.name for p in sub_plugins)}"
            )
        logger.info(plugin_str)

        # Get sorted plugins
        sorted_plugins: Sequence[HHDPlugin] = []
        for plugs in plugins.values():
            sorted_plugins.extend(plugs)
        sorted_plugins.sort(key=lambda x: x.priority)
        validator: Validator = lambda tags, config, value: any(
            p.validate(tags, config, value) for p in sorted_plugins
        )

        if not sorted_plugins:
            logger.error(f"No plugins started, exiting...")
            return

        # Load locales
        locales = []
        for register in pkg_resources.iter_entry_points("hhd.i18n"):
            locales.extend(register.resolve()())
        locales.sort(key=lambda x: x["priority"], reverse=True)

        if locales:
            lstr = "Loaded the following locales:\n"
            for locale in locales:
                lstr += (
                    f" - {locale['domain']} ({locale['priority']}): {locale['dir']}\n"
                )
            logger.info(lstr[:-1])
        else:
            logger.info("No locales found.")

        settings_changed = TEvent()

        def run_plugin_cmd(cmd: Callable[[HHDPlugin], None], reverse: bool = False):
            failed_plugins = []

            arr = enumerate(sorted_plugins)
            if reverse:
                arr = reversed(list(arr))

            for i, p in arr:
                set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
                try:
                    cmd(p)
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    if DEBUG_MODE:
                        raise e
                    logger.error(
                        f"Error while running plugin '{p.name}'. Removing it and continuing..."
                    )
                    failed_plugins.append(i)
                    try:
                        from rich import get_console

                        get_console().print_exception()
                    except Exception:
                        import traceback

                        traceback.print_exc()
                update_log_plugins()

            for i in reversed(failed_plugins):
                del sorted_plugins[i]

            if failed_plugins:
                settings_changed.set()

        # Open plugins
        lock = RLock()
        cond = Condition(lock)
        emit = EmitHolder(cond, ctx, info)
        run_plugin_cmd(lambda p: p.open(emit, ctx))  # type: ignore
        set_log_plugin("main")

        # Compile initial configuration
        state_fn = join(CONFIG_DIR, "state.yml")
        token_fn = join(CONFIG_DIR, ".token")
        settings: HHDSettings = {}
        shash = None

        # Load profiles
        profiles = {}
        templates = {}
        conf = Config({})
        profile_dir = join(CONFIG_DIR, "profiles")
        os.makedirs(profile_dir, exist_ok=True)

        # Monitor config files for changes
        should_initialize = TEvent()
        initial_run = True
        reset = False
        should_exit = TEvent()
        signal.signal(signal.SIGPOLL, notifier(should_initialize, cond))
        signal.signal(signal.SIGINT, notifier(should_exit, cond))
        signal.signal(signal.SIGTERM, notifier(should_exit, cond))

        # Get wakeup count for sleep detection
        wakeup_count = get_wakeup_count()

        while not should_exit.is_set():
            #
            # Configuration
            #

            # Initialize if files changed
            if should_initialize.is_set() or initial_run:
                # wait a bit to allow other processes to save files
                if not initial_run:
                    sleep(INIT_DELAY)
                initial_run = False
                set_log_plugin("main")
                logger.info(f"Reloading configuration.")

                # Settings
                hhd_settings = {"hhd": load_relative_yaml("settings.yml")}
                # TODO: Improve check
                try:
                    if "venv" not in exe_python:
                        del hhd_settings["hhd"]["settings"]["children"]["update_stable"]
                        del hhd_settings["hhd"]["settings"]["children"]["update_beta"]
                except Exception as e:
                    logger.warning(f"Could not hide update settings. Error:\n{e}")

                tmp = [hhd_settings]
                run_plugin_cmd(lambda p: tmp.append(p.settings()))  # type: ignore
                settings = merge_settings(tmp)

                # Force general settings to be last
                if "hhd" in settings:
                    settings = dict(settings)
                    tmp = settings.pop("hhd")
                    settings["hhd"] = tmp
                shash = get_settings_hash(settings)

                # State
                if reset:
                    logger.warning(f"Resetting settings.")
                    conf = get_default_state(settings)
                    conf.updated = True
                    reset = False
                else:
                    new_conf = load_state_yaml(state_fn, settings)
                    if not new_conf:
                        if conf.conf:
                            logger.warning(f"Using previous configuration.")
                        else:
                            logger.info(f"Using default configuration.")
                            conf = get_default_state(settings)
                    else:
                        conf = new_conf

                from importlib.metadata import version

                try:
                    ver = version("hhd")
                    conf["hhd.settings.version"] = ver
                    logger.info(f"Handheld Daemon Version: {ver}")
                except Exception:
                    pass

                try:
                    from hhd.plugins.overlay.overlay import (
                        find_overlay_exe,
                        get_overlay_version,
                    )

                    exe = find_overlay_exe(ctx)
                    if exe:
                        ver = get_overlay_version(exe)
                        conf["hhd.settings.version_ui"] = ver
                        logger.info(f"Overlay Version: {ver}")
                    else:
                        conf["hhd.settings.version_ui"] = "Not Installed"
                        logger.info(f"Overlay not installed")
                except Exception:
                    logger.info(exe)

                # Profiles
                profiles = {}
                templates = {}
                os.makedirs(profile_dir, exist_ok=True)
                for fn in os.listdir(profile_dir):
                    if not fn.endswith(".yml"):
                        continue
                    name = fn.replace(".yml", "")
                    s = load_profile_yaml(join(profile_dir, fn))
                    if s:
                        validate_config(s, settings, validator, use_defaults=False)
                        if name.startswith("_"):
                            templates[name] = s
                        else:
                            # Profiles are shared so lock when accessing
                            # Configs have their own locks and are safe
                            with lock:
                                profiles[name] = s
                if profiles:
                    logger.info(
                        f"Loaded the following profiles (and state):\n[{', '.join(profiles)}]"
                    )
                else:
                    logger.info(f"No profiles found.")

                # Monitor files for changes
                for fd in cfg_fds:
                    try:
                        fcntl.fcntl(fd, fcntl.F_NOTIFY, 0)
                        os.close(fd)
                    except Exception:
                        pass
                cfg_fds = []
                cfg_fns = [
                    CONFIG_DIR,
                    join(CONFIG_DIR, "profiles"),
                ]
                for fn in cfg_fns:
                    fd = -1
                    try:
                        fd = os.open(fn, os.O_RDONLY)
                        fcntl.fcntl(
                            fd,
                            fcntl.F_NOTIFY,
                            fcntl.DN_CREATE
                            | fcntl.DN_DELETE
                            | fcntl.DN_MODIFY
                            | fcntl.DN_RENAME
                            | fcntl.DN_MULTISHOT,
                        )
                    except Exception:
                        if fd != -1:
                            os.close(fd)
                        continue
                    cfg_fds.append(fd)

                should_initialize.clear()
                logger.info(f"Initialization Complete!")

            # Initialize http server
            http_cfg = conf["hhd.http"]
            if http_cfg != prev_http_cfg:
                prev_http_cfg = http_cfg
                if https:
                    https.close()

                from .http import HHDHTTPServer

                port = 5335
                localhost = http_cfg["localhost"].to(bool)
                use_token = http_cfg["token"].to(bool)

                if use_token:
                    # Generate security token
                    if not os.path.isfile(token_fn):
                        import hashlib
                        import random

                        token = hashlib.sha256(
                            str(random.random()).encode()
                        ).hexdigest()[:12]
                        with open(token_fn, "w") as f:
                            os.chmod(token_fn, 0o644)
                            f.write(token)
                    else:
                        with open(token_fn, "r") as f:
                            token = f.read().strip()

                    # Save the token to an ephemeral dir
                    os.makedirs(os.path.dirname(TOKEN_FILE), exist_ok=True)
                    with open(TOKEN_FILE, "w") as f:
                        os.chmod(TOKEN_FILE, 0o644)
                        f.write(token)
                else:
                    token = None

                set_log_plugin("rest")
                https = HHDHTTPServer(localhost, port, token)
                https.update(settings, conf, info, profiles, emit, locales)
                try:
                    https.open()
                except Exception as e:
                    logger.error(
                        f"Could not start http API on port {port}.\n"
                        + "Is another version of Handheld Daemon open?\n"
                        + "Closing."
                    )
                    return
                update_log_plugins()
                set_log_plugin("main")

            #
            # Plugin loop
            #

            # Process events
            set_log_plugin("main")
            events = emit.get_events()

            new_wakeup_count = get_wakeup_count()
            curr = time.time()
            # Debounce sleep event to avoid spurious wakeup triggers
            # This loop will run every 2 seconds, perhaps 4 seconds if there is
            # a delay. Unless 8 seconds lapse, ignore the event
            if (
                new_wakeup_count != wakeup_count
                and last_event
                and curr > last_event + SLEEP_MIN_T
            ):
                logger.info(
                    f"System woke up from sleep. Wakeup count: {new_wakeup_count} from {wakeup_count}."
                )
                events: Sequence[Event] = [
                    *events,
                    {
                        "type": "special",
                        "event": "wakeup",
                        "data": {
                            "count": new_wakeup_count
                        },  # FIXME: Count might be removed in the future
                    },
                ]
            wakeup_count = new_wakeup_count
            last_event = curr

            # AC status
            if ac_fn:
                new_status = get_ac_status(ac_fn)
                if new_status != ac_status:
                    logger.info(f"AC status is: {new_status}")
                    ac_status = new_status
                    info["ac"] = ac_status

            for ev in events:
                match ev["type"]:
                    case "settings":
                        settings_changed.set()
                    case "profile":
                        new_conf = ev["config"]
                        if new_conf:
                            with lock:
                                profiles[ev["name"]] = ev["config"]
                            validate_config(
                                profiles[ev["name"]],
                                settings,
                                validator,
                                use_defaults=False,
                            )
                        else:
                            with lock:
                                if ev["name"] in profiles:
                                    del profiles[ev["name"]]
                    case "apply":
                        if ev["name"] in profiles:
                            conf.update(profiles[ev["name"]].conf)
                    case "state":
                        conf.update(ev["config"].conf)
                    case "special":
                        if ev["event"] == "restart_dev":
                            should_exit.set()
                            switch_root = True
                            break
                        elif ev["event"] == "shutdown_dev":
                            should_exit.set()
                            # Trigger restart
                            updated = True
                    case "acpi" | "tdp" | "ppd" | "energy":
                        pass
                    case other:
                        logger.error(f"Invalid event type submitted: '{other}'")

            # Refresh steam status
            # TODO: Move this?
            refresh_is_steam_running()

            # If settings changed, the configuration needs to reload
            # but it needs to be saved first
            if settings_changed.is_set():
                settings_changed.clear()
                logger.info(f"Reloading settings.")

                # Settings
                settings_base = {
                    k: {} for k in load_relative_yaml("sections.yml")["sections"]
                }
                hhd_settings = {"hhd": load_relative_yaml("settings.yml")}
                # TODO: Improve check
                try:
                    if "venv" not in exe_python:
                        del hhd_settings["hhd"]["settings"]["children"]["update_stable"]
                        del hhd_settings["hhd"]["settings"]["children"]["update_beta"]
                except Exception as e:
                    logger.warning(f"Could not hide update settings. Error:\n{e}")

                tmp = [settings_base, hhd_settings]
                run_plugin_cmd(lambda p: tmp.append(p.settings()))  # type: ignore
                settings = merge_settings(tmp)

                # Force general settings to be last
                if "hhd" in settings:
                    settings = dict(settings)
                    tmp = settings.pop("hhd")
                    settings["hhd"] = tmp
                shash = get_settings_hash(settings)

                # Add new defaults
                conf = Config([parse_defaults(settings), conf.conf])
                conf.updated = True

            # Validate config
            validate_config(conf, settings, validator)

            #
            # Plugin event loop
            #

            # Allow plugins to process events
            if events:
                run_plugin_cmd(lambda p: p.notify(events))

            # Run prepare loop
            run_plugin_cmd(lambda p: p.prepare(conf), reverse=True)

            # Run update loop
            run_plugin_cmd(lambda p: p.update(conf))

            set_log_plugin("ukwn")

            # Notify that events were applied
            # Before saving to reduce delay (yaml files take 100ms :( )
            if https:
                https.update(settings, conf, info, profiles, emit, locales)

            #
            # Save loop
            #

            has_new = should_initialize.is_set()
            saved = False
            # Save existing profiles if open
            if save_state_yaml(state_fn, settings, conf, shash):
                saved = True
                conf.updated = False
            for name, prof in profiles.items():
                fn = join(profile_dir, name + ".yml")
                if save_profile_yaml(fn, settings, prof, shash):
                    saved = True
                    prof.updated = False
            for prof in os.listdir(profile_dir):
                if prof.startswith("_") or not prof.endswith(".yml"):
                    continue
                name = prof[:-4]
                if name not in profiles:
                    fn = join(profile_dir, prof)
                    try:
                        new_fn = fn + ".bak"
                        os.rename(fn, new_fn)
                        saved = True
                    except Exception as e:
                        logger.error(
                            f"Failed removing profile {name} at:\n{fn}\nWith error:\n{e}"
                        )

            if not has_new and saved:
                # We triggered the interrupt, clear
                should_initialize.clear()

            upd_stable = conf.get("hhd.settings.update_stable", False)
            upd_beta = conf.get("hhd.settings.update_beta", False)

            if upd_stable or upd_beta:
                set_log_plugin("main")
                conf["hhd.settings.update_stable"] = False
                conf["hhd.settings.update_beta"] = False

                if ctx:
                    switch_priviledge(ctx.uid)
                try:
                    logger.info(f"Updating Handheld Daemon.")
                    if "venv" in exe_python:
                        subprocess.check_call(
                            [
                                exe_python,
                                "-m",
                                "pip",
                                "uninstall",
                                "-y",
                                "hhd",
                                "adjustor",
                            ]
                        )
                        subprocess.check_call(
                            [
                                exe_python,
                                "-m",
                                "pip",
                                "install",
                                "--upgrade",
                                "--cache-dir",
                                "/tmp/__hhd_update_cache",
                                (GIT_HHD if upd_beta else "hhd"),
                            ]
                        )

                        if not upd_beta:
                            # No beta version for the UI yet, skip updating it
                            import json
                            import stat
                            import urllib.request

                            with urllib.request.urlopen(
                                "https://api.github.com/repos/hhd-dev/hhd-ui/releases/latest"
                            ) as f:
                                release_data = json.load(f)

                            for asset in release_data["assets"]:
                                os.makedirs(
                                    expanduser("~/.local/bin", ctx), exist_ok=True
                                )
                                if "hhd-ui.AppImage" == asset["name"]:
                                    REPORT_POINTS = 8

                                    def progress(idx, blockSize, total):
                                        interval = int(
                                            total / blockSize / REPORT_POINTS
                                        )
                                        if idx % interval == 0:
                                            logger.info(
                                                f"Downloading overlay: {100*idx*blockSize / total:.1f}%"
                                            )

                                    out_fn = expanduser("~/.local/bin/hhd-ui", ctx)
                                    urllib.request.urlretrieve(
                                        asset["browser_download_url"],
                                        out_fn,
                                        reporthook=progress,
                                    )

                                    st = os.stat(out_fn)
                                    os.chmod(out_fn, st.st_mode | stat.S_IEXEC)
                                    break

                        # Set updated
                        updated = True
                    else:
                        logger.error(
                            f"Could not update, python executable is not within a venv (checked for 'venv' in path name):\n{exe_python}"
                        )
                except Exception as e:
                    err = f"Error while updating:\n{e}"
                    conf["hhd.settings.update_error"] = err
                    logger.error(err)
                switch_priviledge(0)

                if updated:
                    should_exit.set()

            # Wait for events
            with lock:
                if (
                    not should_exit.is_set()
                    and not settings_changed.is_set()
                    and not should_initialize.is_set()
                    and not emit.has_events()
                ):
                    cond.wait(timeout=POLL_DELAY)

            # Check reset
            if conf["hhd.settings.reset"].to(bool):
                conf["hhd.settings.reset"] = False
                should_initialize.set()
                reset = True

        set_log_plugin("main")
        logger.info(f"Received interrupt or updated. Stopping plugins and exiting.")
    finally:
        for fd in cfg_fds:
            try:
                os.close(fd)
            except Exception:
                pass
        if https:
            set_log_plugin("main")
            logger.info("Shutting down the REST API.")
            https.close()
        for plugs in plugins.values():
            for p in plugs:
                set_log_plugin("main")
                logger.info(f"Stopping plugin `{p.name}`.")
                set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
                p.close()

        set_log_plugin("main")
        try:
            logger.info("Closing cached controllers.")
            from hhd.controller.virtual.dualsense import Dualsense
            from hhd.controller.virtual.sd import SteamdeckController
            from hhd.controller.virtual.uinput import UInputDevice

            UInputDevice.close_cached()
            Dualsense.close_cached()
            SteamdeckController.close_cached()
        except Exception as e:
            logger.error("Could not close cached controllers with error:\n{e}")

    if updated:
        # Use error code to restart service
        sys.exit(-1)

    if switch_root:
        os.environ["HHD_SWITCH_ROOT"] = "1"
        o = subprocess.run([HHD_DEV_DIR + "/bin/hhd", *sys.argv], check=False)
        sys.exit(o.returncode)


if __name__ == "__main__":
    main()



import logging
import os
import subprocess
import sys
from threading import Thread
from typing import Literal
import shutil
import time
import signal

from hhd.plugins import Context, HHDPlugin, load_relative_yaml
from hhd.plugins.conf import Config

from adjustor.fuse.gpu import (
    get_igpu_status,
    set_cpu_boost,
    set_epp_mode,
    set_gpu_auto,
    set_gpu_manual,
    set_powersave_governor,
    can_use_nonlinear,
    set_frequency_scaling,
)

logger = logging.getLogger(__name__)

APPLY_DELAY = 0.25
SLEEP_DELAY = 4.5


def _ppd_client(emit, proc):
    os.set_blocking(proc.stdin.fileno(), False)

    while True:
        if proc.poll() is not None:
            break
        line = proc.stdout.readline().decode().strip()
        if not line:
            break
        if line not in ("power", "balanced", "performance"):
            logger.error(f"Invalid PPD mode: {line}")
            continue
        emit({"type": "ppd", "status": line})


def _open_ppd_server(emit):
    logger.info("Launching PPD server.")
    proc = subprocess.Popen(
        [sys.executable, "-m", "adjustor.drivers.gpu.ppd"],
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )
    t = Thread(target=_ppd_client, args=(emit, proc))
    t.start()
    return proc, t


class GpuPlugin(HHDPlugin):

    def __init__(
        self,
    ) -> None:
        self.name = f"adjustor_ppd"
        self.priority = 8
        self.log = "agpu"
        self.core_available = False
        self.core_enabled = False
        self.enabled = False
        self.ppd_conflict = False
        self.initialized = False
        self.supports_boost = False
        self.supports_sched = False
        self.supports_freq = False
        self.min_freq = None
        self.avail_scheds = {}

        self.proc = None
        self.t = None

        self.queue = None
        self.queue_gpu = None
        self.sched_proc = None
        self.old_ppd = False
        self.old_freq = None
        self.old_sched = "disabled"
        self.old_boost = None
        self.old_epp = None
        self.old_target = None
        self.old_min_freq = None
        self.gpu_freq_min = None
        self.gpu_freq_max = None
        self.gpu_freq_next = None
        self.target: Literal["power", "balanced", "performance"] = "balanced"

        self.logged_boost = False
        self.logged_error = False

    def settings(self):
        if not self.core_enabled:
            self.initialized = False
            self.core_available = False
            return {}

        try:
            status = get_igpu_status()
        except Exception as e:
            logger.error(f"Failed to get AMD GPU status:\n{e}")

            import traceback

            logger.error(traceback.format_exc())
            status = None

        if not status:
            self.core_available = False
            if not self.logged_error:
                logger.error(
                    "Could not get frequency status. Disabling AMD GPU plugin."
                )
                self.logged_error = True
            return {}

        sets = load_relative_yaml("./settings.yml")
        self.core_available = True
        if not self.enabled:
            self.initialized = False
            return {"hhd": {"settings": sets["core"]}}

        self.ppd_conflict = False
        try:
            out = subprocess.check_output(
                [
                    "systemctl",
                    "list-units",
                    "-t",
                    "service",
                    "--full",
                    # "--all",
                    "--plain",
                    "--no-legend",
                ]
            )
            for line in out.decode().splitlines():
                line = line.lower()
                if "masked" in line:
                    continue
                if "not-found" in line:
                    continue
                if "inactive" in line:
                    continue
                if "power-profiles-daemon" in line or "tuned" in line:
                    self.ppd_conflict = True
                    break
        except Exception as e:
            logger.error(f"Failed to check for PPD conflict:\n{e}")

        if self.ppd_conflict and os.environ.get("HHD_PPD_MASK", None):
            logger.warning(
                "PPD conflict detected but HHD_PPD_MASK is set. Masking PPD/TuneD."
            )
            # Mask and disable
            os.system("systemctl mask power-profiles-daemon.service")
            os.system("systemctl disable --now power-profiles-daemon.service")
            os.system("systemctl mask tuned.service")
            os.system("systemctl disable --now tuned.service")
            # Keep going without check to avoid obscure errors
            self.ppd_conflict = False

        if self.ppd_conflict:
            self.initialized = False
            return {
                "tdp": {"amd_energy": sets["conflict"]},
                "hhd": {"settings": sets["core"]},
            }

        self.initialized = True

        if status.freq_min is not None and status.freq_max is not None:
            self.gpu_freq_min = status.freq_min
            self.gpu_freq_max = status.freq_max
            self.supports_freq = True
            # Initialize frequency settings
            manual_freq = sets["enabled"]["children"]["gpu_freq"]["modes"]["manual"][
                "children"
            ]["frequency"]
            upper_freq = sets["enabled"]["children"]["gpu_freq"]["modes"]["upper"][
                "children"
            ]["frequency"]
            min_freq = sets["enabled"]["children"]["gpu_freq"]["modes"]["range"][
                "children"
            ]["min"]
            max_freq = sets["enabled"]["children"]["gpu_freq"]["modes"]["range"][
                "children"
            ]["max"]

            manual_freq["default"] = ((status.freq_min + status.freq_max) // 200) * 100
            upper_freq["default"] = status.freq_max
            min_freq["default"] = status.freq_min
            max_freq["default"] = status.freq_max
            for freq in (manual_freq, min_freq, max_freq, upper_freq):
                freq["min"] = status.freq_min
                freq["max"] = status.freq_max
            self.min_freq = status.freq_min
        else:
            self.supports_freq = False
            del sets["enabled"]["children"]["gpu_freq"]

        self.supports_boost = status.cpu_boost is not None
        if self.supports_boost:
            if not self.logged_boost:
                logger.info(f"CPU Boost toggling is supported.")
        else:
            if not self.logged_boost:
                logger.warning(f"CPU Boost toggling is not supported.")
            del sets["enabled"]["children"]["mode"]["modes"]["manual"]["children"][
                "cpu_boost"
            ]

        self.supports_nonlinear = can_use_nonlinear()
        if not self.supports_nonlinear:
            del sets["enabled"]["children"]["mode"]["modes"]["manual"]["children"][
                "cpu_min_freq"
            ]

        self.supports_epp = status.epp_avail is not None
        if self.supports_epp:
            epp = sets["enabled"]["children"]["mode"]["modes"]["manual"]["children"][
                "cpu_pref"
            ]
            epp["options"] = {
                k: v for k, v in epp["options"].items() if k in status.epp_avail
            }
        else:
            del sets["enabled"]["children"]["mode"]["modes"]["manual"]["children"][
                "cpu_pref"
            ]

        self.avail_scheds = {}
        avail_pretty = {}
        kernel_supports = os.path.isfile("/sys/kernel/sched_ext/state")
        if kernel_supports:
            for sched, pretty in sets["enabled"]["children"]["mode"]["modes"]["manual"][
                "children"
            ]["sched"]["options"].items():
                if sched == "disabled":
                    avail_pretty[sched] = pretty
                    continue

                exe = shutil.which(sched)
                if exe:
                    self.avail_scheds[sched] = exe
                    avail_pretty[sched] = pretty

        if self.avail_scheds:
            sets["enabled"]["children"]["mode"]["modes"]["manual"]["children"]["sched"][
                "options"
            ] = avail_pretty
        else:
            del sets["enabled"]["children"]["mode"]["modes"]["manual"]["children"][
                "sched"
            ]

        self.logged_boost = True
        return {
            "tdp": {"amd_energy": sets["enabled"]},
            "hhd": {"settings": sets["core"]},
        }

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit

    def notify(self, events):
        for ev in events:
            if ev["type"] == "gpu":
                self.gpu_freq_next = (ev["min"], ev["max"])
            if ev["type"] == "energy":
                self.target = ev["status"]
                try:
                    if self.proc and self.proc.stdin:
                        self.proc.stdin.write(f"{self.target}\n".encode())
                        self.proc.stdin.flush()
                except Exception as e:
                    logger.error(f"Failed to send PPD mode:\n{e}")
                    self.close_ppd()
            elif ev["type"] == "special" and ev.get("event", None) == "wakeup":
                logger.info(
                    f"Waking up from sleep, resetting gpu settings after {SLEEP_DELAY:.1f} seconds."
                )
                self.queue = time.perf_counter() + SLEEP_DELAY

    def update(self, conf: Config):
        self.core_enabled = conf["hhd.settings.tdp_ready"].to(bool)
        if not self.core_enabled or not self.core_available:
            return

        enabled = conf["hhd.settings.amd_energy_enable"].to(bool)
        if enabled != self.enabled:
            self.emit({"type": "settings"})
        self.enabled = enabled

        if self.ppd_conflict and conf.get("tdp.amd_energy.enable", False):
            conf["tdp.amd_energy.enable"] = False
            conf["hhd.steamos.gpu_status"] = "conflict"
            self.emit({"type": "settings"})

        if not self.initialized:
            conf["hhd.steamos.gpu_status"] = "disabled"
            return

        new_ppd = conf["hhd.settings.amd_energy_ppd"].to(bool)
        if new_ppd != self.old_ppd:
            self.old_ppd = new_ppd
            if new_ppd:
                try:
                    self.proc, self.t = _open_ppd_server(self.emit)
                    # Fixup target in case it came before
                    if self.proc.stdin and self.target:
                        self.proc.stdin.write(f"{self.target}\n".encode())
                        self.proc.stdin.flush()
                except Exception as e:
                    logger.error(f"Failed to open PPD server:\n{e}")
                    self.close_ppd()
            else:
                self.close_ppd()

        curr = time.perf_counter()
        queue = self.queue is not None and curr >= self.queue
        if queue:
            self.queue = None

        if conf["tdp.amd_energy.mode.mode"].to(str) == "auto":
            if self.target != self.old_target:
                self.old_target = self.target
                self.queue = curr + APPLY_DELAY

            if queue:
                logger.info(
                    f"Handling energy settings for power profile '{self.target}'."
                )
                try:
                    match self.target:
                        case "balanced":
                            if self.supports_epp:
                                set_powersave_governor()
                                set_epp_mode("balance_power")
                            if self.supports_boost:
                                set_cpu_boost(True)
                            set_frequency_scaling(nonlinear=False)
                        case "performance":
                            if self.supports_epp:
                                set_powersave_governor()
                                set_epp_mode("balance_power")
                            if self.supports_boost:
                                set_cpu_boost(True)
                            set_frequency_scaling(nonlinear=False)
                        case _:  # power
                            if self.supports_epp:
                                set_powersave_governor()
                                set_epp_mode("power")
                            if self.supports_boost:
                                set_cpu_boost(False)
                            set_frequency_scaling(nonlinear=False)
                except Exception as e:
                    logger.error(f"Failed to set energy mode:\n{e}")

            # Unless it is set manually, use the default scheduler.
            self.close_sched()
            self.old_sched = None
            self.old_boost = None
            self.old_epp = None
            self.old_min_freq = None
        else:
            self.old_target = None
            if self.supports_boost:
                new_boost = conf["tdp.amd_energy.mode.manual.cpu_boost"].to(bool)
                if new_boost != self.old_boost or queue:
                    self.old_boost = new_boost
                    try:
                        set_cpu_boost(new_boost == "enabled")
                        set_frequency_scaling(
                            nonlinear=self.supports_nonlinear
                            and conf.get(
                                "tdp.amd_energy.mode.manual.cpu_min_freq", None
                            )
                            == "nonlinear"
                        )
                    except Exception as e:
                        logger.error(f"Failed to set CPU boost:\n{e}")

            if self.supports_epp:
                new_epp = conf["tdp.amd_energy.mode.manual.cpu_pref"].to(str)
                if new_epp != self.old_epp or queue:
                    self.old_epp = new_epp
                    try:
                        # Set governor to powersave as well
                        set_powersave_governor()
                        set_epp_mode(new_epp)  # type: ignore
                    except Exception as e:
                        logger.error(f"Failed to set EPP mode:\n{e}")

            if self.supports_nonlinear:
                new_min_freq = conf["tdp.amd_energy.mode.manual.cpu_min_freq"].to(str)
                if new_min_freq != self.old_min_freq or queue:
                    self.old_min_freq = new_min_freq
                    try:
                        set_frequency_scaling(nonlinear=new_min_freq == "nonlinear")
                    except Exception as e:
                        logger.error(f"Failed to set minimum CPU frequency:\n{e}")

            if self.avail_scheds:
                # Check health and print error
                if self.sched_proc and self.sched_proc.poll():
                    err = self.sched_proc.poll()
                    self.sched_proc = None
                    logger.error(
                        f"Scheduler from sched_ext '{self.old_sched}' closed with error code: {err}"
                    )

                new_sched = conf.get("tdp.amd_energy.mode.manual.sched", "disabled")
                if new_sched != self.old_sched:
                    self.close_sched()
                    self.old_sched = new_sched
                    if new_sched and new_sched != "disabled":
                        logger.info(f"Starting sched_ext scheduler '{new_sched}'")
                        self.sched_proc = subprocess.Popen(
                            self.avail_scheds[new_sched],
                            stderr=subprocess.DEVNULL,
                            stdout=subprocess.DEVNULL,
                        )

        if self.supports_freq:
            # Apply GPU settings
            conf["hhd.steamos.gpu_status"] = "enabled"
            conf["hhd.steamos.gpu_min"] = self.gpu_freq_min
            conf["hhd.steamos.gpu_max"] = self.gpu_freq_max
            new_gpu = conf["tdp.amd_energy.gpu_freq.mode"].to(str)

            # Set gpu frequency and if we switch mode remember it
            if self.gpu_freq_next is not None:
                new_min, new_max = self.gpu_freq_next

                if new_min == None and new_max == None:
                    if conf.get("hhd.steamos.gpu_set", False):
                        new_gpu = "auto"
                        conf["tdp.amd_energy.gpu_freq.mode"] = new_gpu
                        conf["hhd.steamos.gpu_set"] = False
                else:
                    conf["hhd.steamos.gpu_set"] = True
                    if (
                        new_gpu == "auto"
                        or (new_gpu != "range" and new_min is not None)
                        or (new_gpu != "upper" and new_min is None)
                    ):
                        if new_min is None:
                            new_gpu = "upper"
                        else:
                            new_gpu = "range"
                        conf["tdp.amd_energy.gpu_freq.mode"] = new_gpu
                    match new_gpu:
                        case "manual":
                            conf["tdp.amd_energy.gpu_freq.manual.frequency"] = new_max
                        case "upper":
                            conf["tdp.amd_energy.gpu_freq.upper.frequency"] = new_max
                        case "range":
                            conf["tdp.amd_energy.gpu_freq.range.min"] = new_min
                            conf["tdp.amd_energy.gpu_freq.range.max"] = new_max

            match new_gpu:
                case "manual":
                    f = conf["tdp.amd_energy.gpu_freq.manual.frequency"].to(int)
                    new_freq = (f, f)
                case "upper":
                    f = conf["tdp.amd_energy.gpu_freq.upper.frequency"].to(int)
                    new_freq = (self.min_freq or f, f)
                case "range":
                    min_f = conf["tdp.amd_energy.gpu_freq.range.min"].to(int)
                    max_f = conf["tdp.amd_energy.gpu_freq.range.max"].to(int)
                    if max_f < min_f:
                        max_f = min_f
                        conf["tdp.amd_energy.gpu_freq.range.max"] = min_f
                    new_freq = (min_f, max_f)
                case _:
                    new_freq = None

            if new_freq != self.old_freq:
                self.old_freq = new_freq
                self.queue_gpu = curr + APPLY_DELAY

                # If the user changed frequency manually, forget
                # steam set state so it does not reset the settings
                if self.gpu_freq_next is None:
                    conf["hhd.steamos.gpu_set"] = False

            self.gpu_freq_next = None

            if (self.queue_gpu is not None and curr >= self.queue_gpu) or queue:
                self.queue_gpu = None
                try:
                    if new_freq:
                        set_gpu_manual(*new_freq)
                    else:
                        set_gpu_auto()
                except Exception as e:
                    logger.error(f"Failed to set GPU mode:\n{e}")

    def close_ppd(self):
        if self.proc is not None:
            self.proc.send_signal(signal.SIGINT)
            self.proc.wait()
            self.proc = None
        if self.t is not None:
            self.t.join()
            self.t = None

    def close_sched(self):
        if self.sched_proc is not None:
            logger.info(f"Closing sched_ext scheduler '{self.old_sched}'.")
            self.sched_proc.send_signal(signal.SIGINT)
            self.sched_proc.wait()
            self.sched_proc = None

    def close(self):
        self.close_ppd()
        self.close_sched()



import logging
import time
from threading import Event as TEvent, Lock, Thread
from typing import Sequence

from hhd.plugins import Context, Event, HHDPlugin, load_relative_yaml
from hhd.plugins.conf import Config

from adjustor.core.alib import AlibParams, DeviceParams, alib
from adjustor.core.fan import fan_worker, get_fan_info
from adjustor.core.platform import get_platform_choices, set_platform_profile
from adjustor.i18n import _

logger = logging.getLogger(__name__)

PP_DELAY = 0.2
APPLY_DELAY = 1
SLEEP_DELAY = 4

DEFAULT_EDGE = {
    40: 30,
    45: 30,
    50: 40,
    55: 45,
    60: 55,
    65: 60,
    70: 70,
    80: 85,
    90: 100,
}
DEFAULT_TCTL = {
    40: 40,
    50: 45,
    60: 50,
    70: 80,
    80: 90,
    90: 100,
    100: 100,
}


class SmuQamPlugin(HHDPlugin):

    def __init__(
        self,
        dev: dict[str, DeviceParams],
        pp_map: list[tuple[str, list[str], int, int]] | None,
        pp_enable: bool = True,
        init_tdp: bool = True,
    ) -> None:
        self.name = f"adjustor_smu_qam"
        self.priority = 7
        self.log = "smuq"
        self.enabled = False
        self.initialized = False
        self.dev = dev
        self.enforce_limits = True
        self.emit = None
        self.old_conf = None
        self.startup = True
        self.queued = None
        self.sys_tdp = False

        self.old_tdp = None
        self.old_boost = None
        self.new_tdp = None
        self.is_set = False
        self.lims = self.dev.get("skin_limit", self.dev.get("stapm_limit", None))

        self.fan_info = None
        self.fan_t = None
        self.fan_should_exit = TEvent()
        self.fan_junction = TEvent()
        self.fan_lock = Lock()
        self.fan_curve = {}
        self.fan_state = {}

        # Workaround for debugging on the legion go
        # Avoids sending SMU commands that will conflict with Lenovo TDP on
        # startup
        self.init_tdp = init_tdp

        self.pp_map = pp_map
        if pp_enable and pp_map:
            self.pps = get_platform_choices() or []
            if not self.pps:
                logger.warning(
                    f"Platform profile map was provided but device does not have platform profiles."
                )
        else:
            self.pps = []

    def settings(self):
        if not self.enabled:
            self.initialized = False
            return {}

        self.initialized = True
        out = {"tdp": {"qam": load_relative_yaml("qam.yml")}}

        # Set device limits based on stapm
        lims = self.lims
        assert (
            lims
        ), f"Device params do not include skin limit or stapm limit to set tdp."

        dmin, smin, default, smax, dmax = lims
        if self.enforce_limits:
            out["tdp"]["qam"]["children"]["tdp"].update(
                {"min": smin, "max": smax, "default": default}
            )
        else:
            out["tdp"]["qam"]["children"]["tdp"].update(
                {"min": dmin, "max": dmax, "default": default}
            )

        if not self.fan_info:
            del out["tdp"]["qam"]["children"]["fan"]
        else:
            base = out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_edge"][
                "children"
            ]["st40"]
            reset = out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_edge"][
                "children"
            ].pop("reset")
            for k, v in DEFAULT_EDGE.items():
                out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_edge"][
                    "children"
                ][f"st{k}"] = {**base, "title": f"{k}C", "default": v}
            out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_edge"]["children"][
                "reset"
            ] = reset
            reset = out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_junction"][
                "children"
            ].pop("reset")
            for k, v in DEFAULT_TCTL.items():
                out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_junction"][
                    "children"
                ][f"st{k}"] = {**base, "title": f"{k}C", "default": v}
            out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_junction"][
                "children"
            ]["reset"] = reset

            if self.fan_info["tctl"] is None:
                del out["tdp"]["qam"]["children"]["fan"]["modes"]["manual_junction"]

        return out

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit
        self.fan_info = get_fan_info()

    def update(self, conf: Config):
        self.enabled = conf["hhd.settings.tdp_ready"].to(bool)
        self.enforce_limits = conf["hhd.settings.enforce_limits"].to(bool)
        if not self.enabled or not self.initialized:
            self.startup = self.init_tdp
            return

        curr = time.time()
        sys_tdp = False
        if self.new_tdp:
            new_tdp = self.new_tdp
            self.new_tdp = None
            sys_tdp = True
            conf["tdp.qam.tdp"] = new_tdp
        else:
            new_tdp = conf["tdp.qam.tdp"].to(int)

        if self.startup and self.lims:
            smin = self.lims.smin
            smax = self.lims.smax

            if smin and new_tdp < smin:
                logger.warning(
                    f"Device TDP ({new_tdp}) too low for startup, adjusting."
                )
                new_tdp = smin
                conf["tdp.qam.tdp"] = smin
            if smax and new_tdp > smax:
                logger.warning(
                    f"Device TDP ({new_tdp}) too low for startup, adjusting."
                )
                new_tdp = smax
                conf["tdp.qam.tdp"] = smax

        new_boost = conf["tdp.qam.boost"].to(bool)
        changed = (
            (new_tdp != self.old_tdp or new_boost != self.old_boost)
            and self.old_tdp is not None
            and self.old_boost is not None
        )
        if changed and not sys_tdp:
            self.sys_tdp = False

        if self.startup or changed:
            self.queued = curr + APPLY_DELAY
            self.is_set = False

            conf["tdp.smu.std.skin_limit"] = new_tdp
            conf["tdp.smu.std.stapm_limit"] = new_tdp

            if self.pp_map:
                pp = ep = self.pp_map[0][0]
                for nep, npps, tdp, target in self.pp_map:
                    if tdp < new_tdp:
                        ep = nep
                        for npp in npps:
                            if npp in self.pps:
                                pp = npp
                if (
                    self.pps
                    and conf.get("tdp.smu.platform_profile", "disabled") != "disabled"
                ):
                    conf["tdp.smu.platform_profile"] = pp
                conf["tdp.smu.energy_policy"] = ep

            if new_boost:
                try:
                    fmax = self.dev["fast_limit"].smax
                    smax = self.dev["stapm_limit"].smax
                    assert fmax and smax

                    conf["tdp.smu.std.fast_limit"] = int(new_tdp * (fmax / smax))
                    conf["tdp.smu.std.slow_limit"] = min(
                        new_tdp + 2, conf["tdp.smu.std.fast_limit"].to(int)
                    )
                except Exception as e:
                    logger.error(f"Setting boost failed with error:\n{e}")
                    conf["tdp.qam.boost"] = False
            else:
                conf["tdp.smu.std.slow_limit"] = new_tdp
                conf["tdp.smu.std.fast_limit"] = new_tdp

        # Show steam message
        if self.sys_tdp:
            conf["tdp.qam.sys_tdp"] = _("Steam is controlling TDP")
        else:
            conf["tdp.qam.sys_tdp"] = ""

        if self.startup or (self.queued and self.queued < curr):
            self.startup = False
            self.queued = None
            conf["tdp.smu.apply"] = True

        self.old_tdp = new_tdp
        self.old_boost = new_boost

        if self.fan_info:
            mode = conf["tdp.qam.fan.mode"].to(str)
            if mode != "disabled":
                with self.fan_lock:
                    if conf[f"tdp.qam.fan.{mode}.reset"].to(bool):
                        conf[f"tdp.qam.fan.{mode}.reset"] = False
                        curve = DEFAULT_EDGE if "edge" in mode else DEFAULT_TCTL
                        for k, v in curve.items():
                            if f"tdp.qam.fan.{mode}.st{k}" in conf:
                                conf[f"tdp.qam.fan.{mode}.st{k}"] = v

                    for k, v in conf[f"tdp.qam.fan.{mode}"].to(dict).items():
                        if not k.startswith("st"):
                            continue
                        self.fan_curve[int(k[2:])] = v / 100
                    if self.fan_state:
                        s = self.fan_state
                        fan_speed = (
                            f"{s['v_curr']*100:.1f}% @ {s['t_target']}C"
                            if s["in_setpoint"]
                            else f"{s['v_curr']*100:.1f}% → {s['v_target']*100:.1f}%"
                        )
                        conf[f"tdp.qam.fan.{mode}.info"] = (
                            f"{fan_speed} ({', '.join(map(str, s['v_rpm']))} RPM)\n"
                            + (
                                f"Tctl: {s['t_junction']:.2f}C, "
                                if s["t_junction"] is not None
                                else ""
                            )
                            + f"Edge: {s['t_edge']:.2f}C\n"
                        )
                    if "junction" in mode:
                        self.fan_junction.set()
                    else:
                        self.fan_junction.clear()

                if not self.fan_t:
                    self.fan_should_exit.clear()
                    self.fan_t = Thread(
                        target=fan_worker,
                        args=(
                            self.fan_info,
                            self.fan_should_exit,
                            self.fan_lock,
                            self.fan_curve,
                            self.fan_state,
                            self.fan_junction,
                        ),
                    )
                    self.fan_t.start()
            else:
                if self.fan_t:
                    self.fan_should_exit.set()
                    self.fan_t.join()
                    self.fan_t = None
                    self.fan_state = {}

    def notify(self, events: Sequence[Event]):
        for ev in events:
            if ev["type"] == "tdp":
                self.sys_tdp = True
                self.new_tdp = ev["tdp"]
                self.sys_tdp = ev["tdp"] is not None

            if ev["type"] == "ppd" and self.pp_map:
                for ep, pps, tdp, target in self.pp_map:
                    if ep == ev["status"]:
                        self.new_tdp = target
                        break
                else:
                    logger.warning(f"Energy profile '{ev['status']}' not found in map.")

            if ev["type"] == "special" and ev.get("event", None) == "wakeup":
                logger.info(
                    f"Waking up from sleep, resetting TDP after {SLEEP_DELAY} seconds."
                )
                self.queued = time.time() + SLEEP_DELAY

    def close(self):
        if self.fan_t:
            self.fan_should_exit.set()
            self.fan_t.join()
            self.fan_t = None
            self.fan_state = {}


class SmuDriverPlugin(HHDPlugin):

    def __init__(
        self,
        dev: dict[str, DeviceParams],
        cpu: dict[str, AlibParams],
        platform_profile: bool = True,
    ) -> None:
        self.name = f"adjustor_smu"
        self.priority = 9
        self.log = "asmu"
        self.enabled = False
        self.initialized = False
        self.enforce_limits = True

        self.dev = dev
        self.cpu = cpu

        self.old_target = None
        self.check_pp = platform_profile
        self.has_pp = False
        self.old_pp = None
        self.old_vals = {}
        self.is_set = False

        for k in dev:
            assert (
                k in cpu
            ), f"Device supports more keys than what is available in its architecture spec. Key '{k}' missing."

    def settings(self):
        if not self.enabled:
            self.initialized = False
            return {}
        self.initialized = True
        out = {
            "tdp": {
                "smu": load_relative_yaml("smu.yml"),
            }
        }

        # Limit platform profile choices or remove
        choices = get_platform_choices()
        if choices and self.check_pp:
            options = out["tdp"]["smu"]["children"]["platform_profile"]["options"]
            for c in list(options):
                if c not in choices and c != "disabled":
                    del options[c]
            self.has_pp = True
        else:
            del out["tdp"]["smu"]["children"]["platform_profile"]
            self.has_pp = False

        # Remove unsupported instructions
        # Add absolute limits based on CPU
        std = out["tdp"]["smu"]["children"]["std"]["children"]
        for k in list(std):
            if k in self.cpu:
                lims = self.cpu[k]
                std[k].update({"min": lims.min, "max": lims.max})
            else:
                del std[k]
        adv = out["tdp"]["smu"]["children"]["std"]["children"]
        for k in list(adv):
            if k in self.cpu and k != "enable":
                lims = self.cpu[k]
                std[k].update({"min": lims.min, "max": lims.max})
            else:
                del adv[k]

        # Set sane defaults based on device
        std = out["tdp"]["smu"]["children"]["std"]["children"]
        for k in list(std):
            if k in self.dev:
                std[k]["default"] = self.dev[k].default
        adv = out["tdp"]["smu"]["children"]["std"]["children"]
        for k in list(adv):
            if k in self.dev and k != "enable":
                adv[k]["default"] = self.dev[k].default

        return out

    def open(
        self,
        emit,
        context: Context,
    ):
        self.emit = emit

    def update(self, conf: Config):
        self.enabled = conf["hhd.settings.tdp_ready"].to(bool)
        self.enforce_limits = conf["hhd.settings.enforce_limits"].to(bool)
        if not self.enabled or not self.initialized:
            return

        if self.enforce_limits:
            for k, v in conf["tdp.smu.std"].to(dict).items():
                if k in self.dev:
                    mmin, mmax = self.dev[k].smin, self.dev[k].smax
                    if v < mmin:
                        conf["tdp.smu.std", k] = mmin
                    if v > mmax:
                        conf["tdp.smu.std", k] = mmax
            for k, v in conf["tdp.smu.adv"].to(dict).items():
                if k in self.dev and k != "enable":
                    mmin, mmax = self.dev[k].smin, self.dev[k].smax
                    if v < mmin:
                        conf["tdp.smu.adv", k] = mmin
                    if v > mmax:
                        conf["tdp.smu.adv", k] = mmax

        new_vals = {}
        for k, v in conf["tdp.smu.std"].to(dict[str, int]).items():
            new_vals[k] = v
        if conf["tdp.smu.adv.enable"].to(bool):
            for k, v in conf["tdp.smu.adv"].to(dict[str, int]).items():
                if k != "enable":
                    new_vals[k] = v

        if set(new_vals.items()) != set(self.old_vals.items()):
            self.is_set = False

        if self.has_pp:
            new_pp = conf["tdp.smu.platform_profile"].to(str)
            if new_pp != self.old_pp and new_pp != "disabled":
                self.is_set = False
            self.old_pp = new_pp

        # Inform ppd instantly to avoid lag in slider
        new_target = conf["tdp.smu.energy_policy"].to(str)
        if new_target != self.old_target:
            self.old_target = new_target
            self.emit({"type": "energy", "status": new_target})  # type: ignore

        if conf["tdp.smu.apply"].to(bool):
            conf["tdp.smu.apply"] = False

            if self.has_pp:
                cpp = conf["tdp.smu.platform_profile"].to(str)
                if cpp != "disabled":
                    set_platform_profile(cpp)
                    time.sleep(PP_DELAY)

            alib(
                new_vals,
                self.cpu,
                limit="device" if self.enforce_limits else "cpu",
                dev=self.dev,
            )
            self.is_set = True

        self.old_vals = new_vals
        if self.is_set:
            conf["tdp.smu.status"] = "Set"
        else:
            conf["tdp.smu.status"] = "Not Set"

    def close(self):
        pass



import logging
import os
logger = logging.getLogger(__name__)

def find_igpu():
    for hw in os.listdir("/sys/class/hwmon"):
        if not hw.startswith("hwmon"):
            continue
        if not os.path.exists(f"/sys/class/hwmon/{hw}/name"):
            continue
        with open(f"/sys/class/hwmon/{hw}/name", 'r') as f:
            if "amdgpu" not in f.read():
                continue

        if not os.path.exists(f"/sys/class/hwmon/{hw}/device"):
            logger.error(f'No device symlink found for "{hw}"')
            continue

        if not os.path.exists(f"/sys/class/hwmon/{hw}/device/local_cpulist"):
            logger.warning(
                f'No local_cpulist found for "{hw}". Assuming it is a dedicated unit.'
            )
            continue

        pth = os.path.realpath(os.path.join("/sys/class/hwmon", hw))
        return pth

    logger.error("No iGPU found. Binding TDP attributes will not be possible.")
    return None



import logging
import os
from typing import Literal, NamedTuple
from typing import Sequence

from adjustor.fuse.utils import find_igpu as find_amd_igpu

logger = logging.getLogger(__name__)
GPU_FREQUENCY_PATH = "device/pp_od_clk_voltage"
GPU_LEVEL_PATH = "device/power_dpm_force_performance_level"

INTEL_GPU_I915_MIN_FREQ_PATH = "device/gt_RPn_freq_mhz"
INTEL_GPU_I915_MAX_FREQ_PATH = "device/gt_RP0_freq_mhz"
INTEL_GPU_I915_MIN_FREQ_PATH_SET = "device/gt_min_freq_mhz"
INTEL_GPU_I915_MAX_FREQ_PATH_SET = "device/gt_max_freq_mhz"

INTEL_GPU_XE_MIN_FREQ_PATH = "device/tile0/gt0/freq0/rpe_freq"
INTEL_GPU_XE_MAX_FREQ_PATH = "device/tile0/gt0/freq0/rpa_freq"
INTEL_GPU_XE_MIN_FREQ_PATH_SET = "device/tile0/gt0/freq0/min_freq"
INTEL_GPU_XE_MAX_FREQ_PATH_SET = "device/tile0/gt0/freq0/max_freq"

CPU_BOOST_PATH = "/sys/devices/system/cpu/amd_pstate/cpb_boost"
INTEL_BOOST_PATH = "/sys/devices/system/cpu/intel_pstate/no_turbo"

CPU_PATH = "/sys/devices/system/cpu/"
CPU_PREFIX = "cpu"
BOOST_FN = "cpufreq/boost"
EPP_AVAILABLE_FN = "cpufreq/energy_performance_available_preferences"
EPP_FN = "cpufreq/energy_performance_preference"
GOVERNOR_FN = "cpufreq/scaling_governor"

CPU_FREQ_DRIVER_MIN_FN = "cpufreq/cpuinfo_min_freq"
CPU_FREQ_DRIVER_MAX_FN = "cpufreq/cpuinfo_max_freq"
CPU_FREQ_NONLINEAR_MIN_FN = "cpufreq/amd_pstate_lowest_nonlinear_freq"
CPU_FREQ_MAX_FN = "cpufreq/scaling_max_freq"
CPU_FREQ_MIN_FN = "cpufreq/scaling_min_freq"

EPP_MODES = ("performance", "balance_performance", "balance_power", "power")
EppStatus = Literal["performance", "balance_performance", "balance_power", "power"]


class GPUStatus(NamedTuple):
    mode: Literal["auto", "manual", "unknown"]
    freq: int | None
    freq_min: int | None
    freq_max: int | None
    cpu_boost: bool | None
    epp_avail: Sequence[EppStatus] | None
    epp: EppStatus | None


def find_intel_igpu():
    for hw in os.listdir("/sys/class/drm"):
        if not hw.startswith("card"):
            continue
        if not os.path.exists(f"/sys/class/drm/{hw}/device/subsystem_vendor"):
            continue
        with open(f"/sys/class/drm/{hw}/device/subsystem_vendor", "r") as f:
            # intel
            if "1462" not in f.read():
                continue

        if not os.path.exists(f"/sys/class/drm/{hw}/device/local_cpulist"):
            logger.warning(
                f'No local_cpulist found for "{hw}". Assuming it is a dedicated unit.'
            )
            continue

        pth = os.path.realpath(os.path.join("/sys/class/drm", hw))
        return pth

    return None

def find_igpu():
    hwmon = find_intel_igpu()
    if not hwmon:
        hwmon = find_amd_igpu()
        if not hwmon:
            return None, False
        intel = False
    else:
        intel = True
    
    return hwmon, intel


def get_igpu_status():
    hwmon, intel = find_igpu()
    if not hwmon:
        return None

    freq_min = None
    freq_max = None
    freq = None
    mode = "unknown"

    epp_avail = None
    epp = None

    if intel:
        freq = None
        is_xe = os.path.exists(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH))
        if is_xe:
            with open(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH), "r") as f:
                freq_min = int(f.read().strip().lower().replace("mhz", ""))
            with open(os.path.join(hwmon, INTEL_GPU_XE_MAX_FREQ_PATH), "r") as f:
                freq_max = int(f.read().strip().lower().replace("mhz", ""))
            with open(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH_SET), "r") as f:
                freq_min_set = int(f.read().strip().lower().replace("mhz", ""))
            with open(os.path.join(hwmon, INTEL_GPU_XE_MAX_FREQ_PATH_SET), "r") as f:
                freq_max_set = int(f.read().strip().lower().replace("mhz", ""))
        is_i915 = os.path.exists(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH))
        if is_i915:
            with open(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH), "r") as f:
                freq_min = int(f.read().strip().lower().replace("mhz", ""))
            with open(os.path.join(hwmon, INTEL_GPU_I915_MAX_FREQ_PATH), "r") as f:
                freq_max = int(f.read().strip().lower().replace("mhz", ""))
            with open(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH_SET), "r") as f:
                freq_min_set = int(f.read().strip().lower().replace("mhz", ""))
            with open(os.path.join(hwmon, INTEL_GPU_I915_MAX_FREQ_PATH_SET), "r") as f:
                freq_max_set = int(f.read().strip().lower().replace("mhz", ""))
        
        if freq_min_set == freq_min and freq_max_set == freq_max:
            mode = "auto"
        else:
            mode = "manual"
    else:
        with open(os.path.join(hwmon, GPU_FREQUENCY_PATH), "r") as f:
            for line in f.readlines():
                if line.startswith("0:"):
                    freq = int(line.split()[1].lower().replace("mhz", ""))
                if line.startswith("SCLK"):
                    freq_min = int(line.split()[1].lower().replace("mhz", ""))
                    freq_max = int(line.split()[2].lower().replace("mhz", ""))

        with open(os.path.join(hwmon, GPU_LEVEL_PATH), "r") as f:
            m = f.read()[:-1]
            if m == "auto":
                mode = "auto"
            elif m == "manual":
                mode = "manual"
            else:
                mode = "unknown"

    cpu_boost_fn = os.path.join(CPU_PATH, CPU_PREFIX + "0", BOOST_FN)
    if os.path.exists(cpu_boost_fn):
        with open(cpu_boost_fn, "r") as f:
            cpu_boost = f.read().strip() == "1"
    elif os.path.exists(CPU_BOOST_PATH):
        with open(CPU_BOOST_PATH, "r") as f:
            cpu_boost = f.read().strip() == "1"
    elif os.path.exists(INTEL_BOOST_PATH):
        with open(INTEL_BOOST_PATH, "r") as f:
            cpu_boost = f.read().strip() == "0"
    else:
        cpu_boost = None

    epp_avail_fn = os.path.join(CPU_PATH, CPU_PREFIX + "0", EPP_AVAILABLE_FN)
    if os.path.exists(epp_avail_fn):
        with open(epp_avail_fn, "r") as f:
            epp_avail: Sequence[EppStatus] | None = [
                p for p in f.read().strip().split() if p in EPP_MODES
            ]

    epp_fn = os.path.join(CPU_PATH, CPU_PREFIX + "0", EPP_FN)
    if os.path.exists(epp_fn):
        with open(epp_fn, "r") as f:
            tmp = f.read().strip().split()
            if tmp in EPP_MODES:
                epp = tmp

    return GPUStatus(
        mode=mode,
        freq=freq,
        freq_min=freq_min,
        freq_max=freq_max,
        cpu_boost=cpu_boost,
        epp_avail=epp_avail,
        epp=epp,
    )


def set_gpu_auto():
    hwmon, intel = find_igpu()
    if not hwmon:
        return None

    if intel:
        logger.info("Setting Intel GPU to auto frequencies.")
        is_xe = os.path.exists(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH))
        if is_xe:
            with open(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH), "r") as f:
                freq_min = f.read().strip().lower().replace("mhz", "")
            with open(os.path.join(hwmon, INTEL_GPU_XE_MAX_FREQ_PATH), "r") as f:
                freq_max = f.read().strip().lower().replace("mhz", "")
            with open(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH_SET), "w") as f:
                f.write(freq_min)
            with open(os.path.join(hwmon, INTEL_GPU_XE_MAX_FREQ_PATH_SET), "w") as f:
                f.write(freq_max)
        is_i915 = os.path.exists(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH))
        if is_i915:
            with open(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH), "r") as f:
                freq_min = f.read().strip().lower().replace("mhz", "")
            with open(os.path.join(hwmon, INTEL_GPU_I915_MAX_FREQ_PATH), "r") as f:
                freq_max = f.read().strip().lower().replace("mhz", "")
            with open(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH_SET), "w") as f:
                f.write(freq_min)
            with open(os.path.join(hwmon, INTEL_GPU_I915_MAX_FREQ_PATH_SET), "w") as f:
                f.write(freq_max)
    else:
        logger.info("Setting GPU mode to 'auto'.")
        hwmon = find_amd_igpu()
        if not hwmon:
            return None
        with open(os.path.join(hwmon, GPU_LEVEL_PATH), "w") as f:
            f.write("auto")


def set_gpu_manual(min_freq: int, max_freq: int | None = None):
    if max_freq is None:
        max_freq = min_freq

    hwmon, intel = find_igpu()
    if not hwmon:
        return None
    
    logger.info(f"Pinning GPU frequency to '{min_freq}Mhz' - '{max_freq}Mhz'.")

    if intel:
        is_xe = os.path.exists(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH))
        if is_xe:
            with open(os.path.join(hwmon, INTEL_GPU_XE_MIN_FREQ_PATH_SET), "w") as f:
                f.write(str(min_freq))
            with open(os.path.join(hwmon, INTEL_GPU_XE_MAX_FREQ_PATH_SET), "w") as f:
                f.write(str(max_freq))
        is_i915 = os.path.exists(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH))
        if is_i915:
            with open(os.path.join(hwmon, INTEL_GPU_I915_MIN_FREQ_PATH_SET), "w") as f:
                f.write(str(min_freq))
            with open(os.path.join(hwmon, INTEL_GPU_I915_MAX_FREQ_PATH_SET), "w") as f:
                f.write(str(max_freq))
    else:
        with open(os.path.join(hwmon, GPU_LEVEL_PATH), "w") as f:
            f.write("manual")

        for cmd in [f"s 0 {min_freq}\n", f"s 1 {max_freq}\n", f"c\n"]:
            with open(os.path.join(hwmon, GPU_FREQUENCY_PATH), "w") as f:
                f.write(cmd)


def read_from_cpu0(fn: str):
    with open(os.path.join(CPU_PATH, CPU_PREFIX + "0", fn), "r") as f:
        return f.read().strip()


def is_in_cpu0(fn: str):
    return os.path.exists(os.path.join(CPU_PATH, CPU_PREFIX + "0", fn))


def set_per_cpu(fn: str, value: str):
    for dir in os.listdir(CPU_PATH):
        if not dir.startswith(CPU_PREFIX):
            continue
        # Make sure CPU# is a number
        try:
            int(dir[len(CPU_PREFIX) :])
        except ValueError:
            continue
        with open(os.path.join(CPU_PATH, dir, fn), "w") as f:
            f.write(value)


def set_cpu_boost(enable: bool):
    logger.info(f"{'Enabling' if enable else 'Disabling'} CPU boost.")
    if os.path.exists(INTEL_BOOST_PATH):
        with open(INTEL_BOOST_PATH, "w") as f:
            f.write("0" if enable else "1")
    elif os.path.exists(CPU_BOOST_PATH):
        try:
            with open(CPU_BOOST_PATH, "w") as f:
                f.write("1" if enable else "0")
        except Exception:
            with open(CPU_BOOST_PATH, "w") as f:
                f.write("enabled" if enable else "disabled")
    elif is_in_cpu0(BOOST_FN):
        set_per_cpu(BOOST_FN, "1" if enable else "0")


def set_epp_mode(mode: EppStatus):
    logger.info(f"Setting EPP mode to '{mode}'.")
    set_per_cpu(EPP_FN, mode)


def set_powersave_governor():
    logger.info("Setting CPU governor to 'powersave'.")
    set_per_cpu(GOVERNOR_FN, "powersave")


def can_use_nonlinear():
    return is_in_cpu0(CPU_FREQ_NONLINEAR_MIN_FN)


def set_frequency_scaling(nonlinear: bool):
    if nonlinear:
        min_freq = read_from_cpu0(CPU_FREQ_NONLINEAR_MIN_FN)
    else:
        min_freq = read_from_cpu0(CPU_FREQ_DRIVER_MIN_FN)
    max_freq = read_from_cpu0(CPU_FREQ_DRIVER_MAX_FN)

    try:
        logger.info(
            f"Setting CPU frequency scaling to [{int(min_freq)/1e6:.3f} GHz, {int(max_freq)/1e6:.3f} GHz]{' (nonlinear)' if nonlinear else ''}."
        )
    except Exception:
        pass
    set_per_cpu(CPU_FREQ_MIN_FN, min_freq)
    set_per_cpu(CPU_FREQ_MAX_FN, max_freq)
